# Copyright (C) 2013-2014 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["parameters"]

# ModelParameter imports
from modelparameters.parameters import Param, OptionParam, ScalarParam
from modelparameters.parameterdict import ParameterDict

parameters = ParameterDict(
    # Generation parameters
    generation=ParameterDict(
        class_code=Param(
            False, description="If true methods are contained " "inside a class"
        ),
        # Code generation parameters
        code=ParameterDict(
            # Float precision
            float_precision=OptionParam(
                "double",
                ["double", "single"],
                description="Float precision in generated " "code.",
            ),
            # Parameter for default argument order
            default_arguments=OptionParam(
                "stp",
                ["tsp", "stp", "spt", "ts", "st"],
                description="Default input argument order: "
                "s=states, p=parameters, t=time",
            ),
            # Parameter for the time parameter name
            time=ParameterDict(name=Param("t", description="Name of time argument")),
            # Parameter for the time step parameter name
            dt=ParameterDict(
                name=Param("dt", description="Name of time step argument")
            ),
            # Number of nodes for solvers
            n_nodes=ScalarParam(0, ge=0, description="Number of nodes for simulation"),
            # Parameters for code generation of arrays
            array=ParameterDict(
                index_format=OptionParam(
                    "[]",
                    ["[]", "{}", "()"],
                    description="The format of index notations.",
                ),
                index_offset=ScalarParam(
                    0, ge=0, description="A global offset to all indexed variables."
                ),
                flatten=Param(
                    True,
                    description="If true multidimensional arrays will be "
                    "flattened. jac[2,3] -> jac[27] if the shape of jac "
                    "is (12,12)",
                ),
            ),
            # Parameters for code generation of parameters
            parameters=ParameterDict(
                field_parameters=Param(
                    [""],
                    description="A list of parameter names "
                    "which should be treated as field parameters. "
                    "Only available for CUDA backends.",
                ),
                representation=OptionParam(
                    "named",
                    ["named", "array", "numerals"],
                    description="Controls how parameters are "
                    "represented in the code. As named variables,"
                    " as an indexed array or as the default "
                    "numeral values given in the gotran model.",
                ),
                array_name=Param(
                    "parameters",
                    description="The name of the array " "representing the parameters.",
                ),
                field_array_name=Param(
                    "field_parameters",
                    description="The name of the array "
                    "representing the field parameters.",
                ),
                add_offset=Param(
                    False,
                    description="If true an offset will be "
                    "added to the index of each parameter",
                ),
                add_field_offset=Param(
                    False,
                    description="If true an offset will be "
                    "added to the index of each field parameter",
                ),
            ),
            # Parameters for code generation of states
            states=ParameterDict(
                field_states=Param(
                    [""],
                    description="A list of state names "
                    "which should be treated as field states. "
                    "Not available for all backends...",
                ),
                representation=OptionParam(
                    "named",
                    ["named", "array"],
                    description="Controls how states are "
                    "represented in the code. As named variables,"
                    " or as an indexed array.",
                ),
                array_name=Param(
                    "states",
                    description="The name of the array " "representing the states.",
                ),
                add_offset=Param(
                    False,
                    description="If true an offset will be "
                    "added to the index of each state",
                ),
            ),
            # Parameters for code generation of body expressions
            body=ParameterDict(
                use_cse=Param(
                    False,
                    description="If true will the body be "
                    "optimized using SymPy common sub expression "
                    "extraction.",
                ),
                use_enum=Param(
                    False,
                    description="If true use enumeration"
                    "data types instead of indexing.",
                ),
                in_signature=Param(
                    False,
                    description="If true the body argument "
                    "will be included in the signature.",
                ),
                representation=OptionParam(
                    "named",
                    ["named", "array", "reused_array"],
                    description="Controls how body variables are "
                    "represented in the code. As named variables,"
                    "as an indexed array or as indexed array with "
                    "reuse of unused array elements.",
                ),
                array_name=Param(
                    "body",
                    description="The name of the array " "representing the body.",
                ),
                optimize_exprs=OptionParam(
                    "none",
                    ["none", "numerals", "numerals_symbols"],
                    description="Remove body expressions as "
                    "intermediates, which contains only a "
                    "numeral or numerals and a symbol.",
                ),
            ),
        ),
        # Parameters for automatic generation of specific functions
        functions=ParameterDict(
            rhs=ParameterDict(
                generate=Param(
                    True,
                    description="Generate code for the "
                    "evaluation of the right hand side evaluation.",
                ),
                function_name=Param(
                    "rhs", description="The name of " "the generated function."
                ),
                result_name=Param(
                    "values", description="The name of " "the result argument."
                ),
            ),
            monitored=ParameterDict(
                generate=Param(
                    True,
                    description="Generate code for the "
                    "evaluation of monitored intermediates.",
                ),
                function_name=Param(
                    "monitor", description="The name of " "the generated function."
                ),
                result_name=Param(
                    "monitored", description="The name of " "the result argument."
                ),
            ),
            jacobian=ParameterDict(
                generate=Param(
                    False,
                    description="Generate code for the "
                    "evaluation of the jacobian of the right hand "
                    "side.",
                ),
                function_name=Param(
                    "compute_jacobian",
                    description="The name " "of the generated function.",
                ),
                result_name=Param(
                    "jac", description="The name of " "the result argument."
                ),
            ),
            lu_factorization=ParameterDict(
                generate=Param(
                    False,
                    description="Generate code for "
                    "symbolicly factorize the jacobian.",
                ),
                function_name=Param(
                    "lu_factorize", description="The name " "of the generated function."
                ),
            ),
            forward_backward_subst=ParameterDict(
                generate=Param(
                    False,
                    description="Generate code for the "
                    "symbolic forward backward substitution of the "
                    "jacobian.",
                ),
                function_name=Param(
                    "forward_backward_subst",
                    description="The name of the generated " "function.",
                ),
                residual_name=Param(
                    "F", description="The name of " "the residual argument."
                ),
                result_name=Param(
                    "dx", description="The name of " "the incriment argument."
                ),
            ),
            componentwise_rhs_evaluation=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "computing componentwise evaluation of the rhs.",
                ),
                function_name=Param(
                    "componentwise_rhs",
                    description="The name of the generated " "function.",
                ),
            ),
            linearized_rhs_evaluation=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "computing linearized evaluation of linear rhs "
                    "terms.",
                ),
                function_name=Param(
                    "linearized_rhs",
                    description="The name of the generated " "function.",
                ),
                include_rhs=Param(
                    False,
                    description="If True the rhs will be "
                    "included as a result argument.",
                ),
                only_linear=Param(
                    True,
                    description="If True only linearized "
                    "expressions for the linear derivatives will "
                    "be generated.",
                ),
                result_names=Param(
                    ["linearized", "rhs"],
                    description="The name of " "the two results arguments.",
                ),
            ),
        ),
        # Parameters for automatic generation of lists/arrays of names for states and parameters
        lists = ParameterDict(
            parameter_names = ParameterDict(
                generate = Param(
                    False,
                    description="Generate list of parameter names"
                ),
                name = Param(
                    "parameter_names",
                    description="Name of list variable"
                ),
            ),
            state_names = ParameterDict(
                generate = Param(
                    False,
                    description="Generate list of state names"
                ),
                name = Param(
                    "state_names",
                    description="Name of list variable"
                ),
            ),
        ),
        # Parameters for automatic generation of specific solver functions
        solvers=ParameterDict(
            explicit_euler=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "solving an ODE using explicit Euler method.",
                ),
                function_name=Param(
                    "forward_explicit_euler",
                    description="The name of the generated " "function.",
                ),
            ),
            rush_larsen=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "solving an ODE using Rush Larsen method.",
                ),
                function_name=Param(
                    "forward_rush_larsen",
                    description="The name of the generated " "function.",
                ),
                delta=ScalarParam(
                    1e-8,
                    gt=0,
                    lt=1.0,
                    description="Value to "
                    "safeguard the evaluation of the rush larsen "
                    "step.",
                ),
            ),
            generalized_rush_larsen=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "solving an ODE using generalized Rush Larsen method.",
                ),
                function_name=Param(
                    "forward_generalized_rush_larsen",
                    description="The name of the generated " "function.",
                ),
                delta=ScalarParam(
                    1e-8,
                    gt=0,
                    lt=1.0,
                    description="Value to "
                    "safeguard the evaluation of the rush larsen "
                    "step.",
                ),
            ),
            hybrid_generalized_rush_larsen=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "solving an ODE using a hybrid generalized Rush Larsen method.",
                ),
                function_name=Param(
                    "forward_hybrid_generalized_rush_larsen",
                    description="The name of the generated " "function.",
                ),
                delta=ScalarParam(
                    1e-8,
                    gt=0,
                    lt=1.0,
                    description="Value to "
                    "safeguard the evaluation of the rush larsen "
                    "step.",
                ),
                stiff_states=Param(
                    "",
                    name="stiff_states",
                    description="List of names of state variables that should "
                    "be solved with the GRL1 scheme that has better stability"
                    "than explicit Euler",
                ),
            ),
            simplified_implicit_euler=ParameterDict(
                generate=Param(
                    False,
                    description="If true, generate code for "
                    "solving an ODE using Rush Larsen method.",
                ),
                function_name=Param(
                    "forward_simplified_implicit_euler",
                    description="The name of the generated " "function.",
                ),
                numeric_jacobian=Param(
                    False, description="Use numeric " "calculated diagonal jacobian."
                ),
            ),
        ),
    ),
    # Parameters for different input
    input=ParameterDict(
        # Parameters for CellML input
        cellml=ParameterDict(
            change_state_names=Param(
                [],
                description="A list of state names "
                "which should be changed to not interfere "
                "with for example a parameter name.",
            ),
            grouping=OptionParam(
                "encapsulation",
                ["encapsulation", "containment"],
                description="Determines what type of grouping "
                "should be used when the cellml model is parsed.",
            ),
            use_sympy_integers=Param(
                False,
                description="If yes dedicated sympy "
                "integers will be used instead of the "
                "integers 0-10. This will turn 1/2 into a "
                "sympy rational instead of the float 0.5.",
            ),
            strip_parent_name=Param(
                True,
                description="If True strip the name from "
                "the child component it contains the name of the "
                "parent component.",
            ),
        )
    ),
)



#!/usr/bin/env python3

import os

from gotran.model.loadmodel import load_ode
from gotran.codegeneration.codegenerators import PythonCodeGenerator
from gotran.common.options import parameters
from gotran.common import error, info, list_timings, Timer
from modelparameters.parameterdict import *


def gotran2py(filename, params):
    """
    Create a c header file from a gotran model
    """
    timer = Timer("Generate Python code from {}".format(filename))

    # Load Gotran model
    ode = load_ode(filename)

    # Collect monitored
    if params.functions.monitored.generate:
        monitored = [expr.name for expr in ode.intermediates + ode.state_expressions]
    else:
        monitored = None

    # Create a Python code generator
    gen = PythonCodeGenerator(
        params,
        ns=params.namespace,
        import_inside_functions=params.import_inside_functions,
    )

    output = params.output

    if output:
        if not output.endswith(".py"):
            output += ".py"
    else:
        output = filename.replace(".ode", "") + ".py"

    info("")
    info("Generating Python code for the {0} ode...".format(ode.name))
    if params.class_code:
        code = gen.class_code(ode, monitored=monitored)
    else:
        code = gen.module_code(ode, monitored=monitored)
    info("  done.")
    with open(output, "w") as f:
        f.write(code)

    del timer

    if params.list_timings:
        list_timings()


def main():
    import sys, os

    generation_params = parameters.generation.copy()

    params = ParameterDict(
        list_timings=Param(
            False,
            description="If true timings for reading "
            "and evaluating the model is listed.",
        ),
        output=Param("", description="Specify output file name"),
        import_inside_functions=Param(
            False, description="Perform imports inside functions"
        ),
        namespace=OptionParam(
            "math",
            ["math", "np", "numpy", "ufl"],
            description="The math namespace of the generated code",
        ),
        **generation_params
    )
    params.parse_args(usage="usage: %prog FILE [options]")  # sys.argv[2:])

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotran2py(file_name, params)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

# Gotran imports
from gotran.input.cellml import CellMLParser


def define_parser(change_state_names):
    """
    Return a populated option parser
    """
    import optparse

    parser = optparse.OptionParser(
        usage="%prog [options]", description="Translate CellML files into gotran files."
    )

    def list_parser(option, opt_str, value, parser, arg_list):
        rargs = parser.rargs
        # print rargs
        while rargs:
            arg = rargs[0]

            # Stop if we hit an arg like "--par", i.e, PAR_PREFIX
            if arg[0] == "-":
                break
            else:
                arg_list.append(rargs.pop(0))

    opt = parser.add_option(
        "-c",
        "--change_state_names",
        help="Change the name locally of a state name. " "Usage: -c R",
        action="callback",
        callback=list_parser,
        callback_args=(change_state_names,),
        metavar="EQN1 EQN",
    )

    parser.add_option(
        "-s",
        "--use_sympy_integers",
        help="If yes dedicated sympy "
        "integers will be used instead of the "
        "integers 0-10. This will turn 1/2 into a "
        "sympy rational instead of the float 0.5.",
        action="store_true",
        default=False,
        dest="use_sympy_integers",
    )

    parser.add_option(
        "-p",
        "--strip_parent_name",
        help="If True strip the name from "
        "the child component it contains the name of the "
        "parent component.",
        action="store_true",
        default=True,
        dest="strip_parent_name",
    )

    parser.add_option(
        "-o",
        "--output",
        help="""Name of output file. If not provided 
        then the same name as the name of the cellml file will be used""",
        default="",
        type=str,
        dest="output",
    )
    return parser


def main():
    import sys
    import os

    change_state_names = []
    parser = define_parser(change_state_names)

    options, args = parser.parse_args()

    param = CellMLParser.default_parameters()

    for name in param:
        if hasattr(options, name):
            param[name] = getattr(options, name)

    if len(args) != 1:
        print(args)
        raise ValueError("Expected 1 and only one CellML file.")

    cellml = CellMLParser(args[0], param)
    gotran_code = cellml.to_gotran()

    if options.output == "":
        output = cellml.name
    else:
        output = os.path.splitext(options.output)[0]

    open("{0}.ode".format(output), "w").write(gotran_code)


if __name__ == "__main__":
    main()



import warnings
from copy import deepcopy
import numpy as np

try:
    import scipy.integrate as spi

    has_scipy = True
except:
    has_scipy = False

from gotran.common import warning
from .odesolver import Solver, ODESolverError

__all__ = ["ScipySolver", "has_scipy"]


class ScipySolver(Solver):
    def __init__(self, ode, **options):

        msg = "Chosen backend is scipy, but scipy is not installed"
        assert has_scipy, msg

        Solver.__init__(self, ode, **options)

        self._options = ScipySolver.list_solver_options()
        self._options.update(
            (k, v) for k, v in options.items() if k in list(self._options.keys())
        )

    def get_options(self):
        return self._options

    # @staticmethod
    # def list_solver_options():
    #     return {'atol': 1e-6,
    #             'max_step': np.inf,
    #             'rtol': 1e-6}

    # These are the old ones
    @staticmethod
    def list_solver_options():
        return {
            "atol": None,
            "col_deriv": 0,
            "full_output": 0,
            "h0": 0.0,
            "hmax": 0.0,
            "hmin": 0.0,
            "ixpr": 0,
            "ml": None,
            "mu": None,
            "mxhnil": 0,
            "mxordn": 12,
            "mxords": 5,
            "mxstep": 0,
            "printmessg": 0,
            "rtol": None,
            "tcrit": None,
        }

    def _solve(self, tsteps, attempts=3):
        """
        Solve ode using scipy.integrade.odeint

        Arguments
        ---------
        tsteps : array
            The time steps
        attempts : int
            If integration fails, i.e the solver does not 
            converge, we could reduce the step size and try again.
            This varible controls how many time we should try to
            solve the problem. Default: 3
        
        """

        # Some flags
        it = 0
        converged = False

        # Get solver options
        options = deepcopy(self._options)

        while it < attempts and not converged:

            # Somehow scipy only display a warning if the ODE itegrator fails.
            # We can record these warnings using the warning module
            with warnings.catch_warnings(record=True) as caught_warnings:

                # Allways catch warnings (not only the first)
                warnings.simplefilter("always")

                fun = lambda y, t: self._rhs(t, y, self._model_params)
                # Solve ode
                results = spi.odeint(fun, self._y0, tsteps, Dfun=self._jac, **options)
                t, y = tsteps, results

                # fun=lambda t, y: self._rhs(t, y, self._model_params)
                # results = spi.solve_ivp(fun=fun,
                #                         y0=self._y0,
                #                         t_span=[tsteps[0], tsteps[-1]],
                #                         method='BDF', #'BDF', 'LSODA' 'Radau'
                #                         t_eval=tsteps,
                #                         jac=self._jac,
                #                         **options)
                # t, y = results.t, results.y

            # Check if we caught any warnings
            converged = len(caught_warnings) == 0
            it += 1
            # If we did, reduce maximum step size
            options["hmax"] /= 2.0

        # If we still caught some warnings raise exception
        if len(caught_warnings) > 0:
            for w in caught_warnings:
                msg = "Catched warning {}\n{}".format(w.category, w.message)
                warning(msg)

                if w.category == spi.odepack.ODEintWarning:
                    raise ODESolverError(msg)

        return t, y



# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    "Expression",
    "DerivativeExpression",
    "AlgebraicExpression",
    "StateExpression",
    "StateSolution",
    "RateExpression",
    "Intermediate",
    "StateDerivative",
    "Derivatives",
    "IndexedExpression",
    "StateIndexedExpression",
    "ParameterIndexedExpression",
    "recreate_expression",
]

# ModelParameters imports
from modelparameters.parameters import SlaveParam
from modelparameters.sympytools import sp
from modelparameters.codegeneration import sympycode, latex
from sympy.core.function import AppliedUndef

# Local imports
from gotran.common import (
    error,
    check_arg,
    scalars,
    debug,
    DEBUG,
    get_log_level,
    Timer,
    tuplewrap,
    parameters,
)
from gotran.model.odeobjects import *


def recreate_expression(expr, *replace_dicts, **kwargs):
    """
    Recreate an Expression while applying replace dicts in given order
    """

    replace_type = kwargs.get("replace_type", "xreplace")
    if not replace_type in ["xreplace", "subs"]:
        error(
            "Valid alternatives for replace_type is: 'xreplace', "
            "'subs' got {0}".format(replace_type)
        )

    # First do the replacements
    sympyexpr = expr.expr
    for replace_dict in replace_dicts:
        if replace_type == "xreplace":
            sympyexpr = sympyexpr.xreplace(replace_dict)
        else:
            sympyexpr = sympyexpr.subs(replace_dict)

    # FIXME: Should we distinguish between the different
    # FIXME: intermediates?
    if isinstance(expr, Intermediate):
        new_expr = Intermediate(expr.name, sympyexpr)

    elif isinstance(expr, StateDerivative):
        new_expr = StateDerivative(expr.state, sympyexpr)

    elif isinstance(expr, AlgebraicExpression):
        new_expr = AlgebraicExpression(expr.state, sympyexpr)

    elif isinstance(expr, IndexedExpression):
        new_expr = IndexedExpression(
            expr.basename,
            expr.indices,
            sympyexpr,
            expr.shape,
            expr._array_params,
            expr._offset_str,
            enum=expr.enum,
        )
    elif isinstance(expr, StateIndexedExpression):
        new_expr = StateIndexedExpression(
            expr.basename,
            expr.indices,
            sympyexpr,
            expr.state,
            expr.shape,
            expr._array_params,
            expr._offset_str,
        )
    elif isinstance(expr, ParameterIndexedExpression):
        new_expr = ParameterIndexedExpression(
            expr.basename,
            expr.indices,
            sympyexpr,
            expr.state,
            expr.parameter,
            expr._array_params,
            expr._offset_str,
        )
    else:
        error("Should not reach here")

    # Inherit the count of the old expression
    new_expr._recount(expr._count)

    return new_expr


class Expression(ODEValueObject):
    """
    class for all expressions such as intermediates and derivatives
    """

    def __init__(self, name, expr, dependent=None):
        """
        Create an Expression with an associated name

        Arguments
        ---------
        name : str
            The name of the Expression
        expr : sympy.Basic
            The expression
        dependent : ODEObject
            If given the count of this Expression will follow as a
            fractional count based on the count of the dependent object
        """

        from modelparameters.sympytools import symbols_from_expr

        # Check arguments
        check_arg(expr, scalars + (sp.Basic,), 1, Expression)

        expr = sp.sympify(expr)

        # Deal with Subs in sympy expression
        for sub_expr in expr.atoms(sp.Subs):

            # deal with one Subs at a time
            subs = dict(
                (key, value) for key, value in zip(sub_expr.variables, sub_expr.point)
            )

            expr = expr.subs(sub_expr, sub_expr.expr.xreplace(subs))

        # Deal with im and re
        im_exprs = expr.atoms(sp.im)
        re_exprs = expr.atoms(sp.re)
        if im_exprs or re_exprs:
            replace_dict = {}
            for im_expr in im_exprs:
                replace_dict[im_expr] = sp.S.Zero
            for re_expr in re_exprs:
                replace_dict[re_expr] = re_expr.args[0]
            expr = expr.xreplace(replace_dict)

        if not symbols_from_expr(expr, include_numbers=True):
            error(
                "expected the expression to contain at least one " "Symbol or Number."
            )

        # Call super class with expression as the "value"
        super(Expression, self).__init__(name, expr, dependent)

        # Collect dependent symbols
        dependent = tuple(
            sorted(
                symbols_from_expr(expr),
                key=cmp_to_key(lambda a, b: cmp(sympycode(a), sympycode(b))),
            )
        )

        if dependent:
            self._sym = self._param.sym(*dependent)
            self._sym._assumptions["real"] = True
            self._sym._assumptions["commutative"] = True
            self._sym._assumptions["imaginary"] = False
            self._sym._assumptions["hermitian"] = True
            self._sym._assumptions["complex"] = True
        else:
            self._sym = self.param.sym

        self._dependent = set(dependent)

    @property
    def dependent(self):
        return self._dependent

    @property
    def expr(self):
        """
        Return the stored expression
        """
        return self._param.expr

    @property
    def sym(self):
        """
        """
        return self._sym

    def replace_expr(self, *replace_dicts):
        """
        Replace registered expression using passed replace_dicts
        """
        expr = self.expr
        for replace_dict in replace_dicts:
            expr = expr.xreplace(replace_dict)
        self._param = SlaveParam(expr, name=self.name)

    @property
    def is_state_expression(self):
        """
        True of expression is a state expression
        """
        return self._is_state_expression

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "'{0}', {1}".format(self.name, sympycode(self.expr))

    def _repr_latex_(self):
        """
        Return a pretty latex representation of the Expression object
        """

        return "${0} = {1}$".format(self._repr_latex_name(), self._repr_latex_expr())

    def _repr_latex_expr(self):
        return latex(self.expr)

    def _repr_latex_name(self):
        return "{0}".format(latex(self.name))


class Intermediate(Expression):
    """
    A class for all Intermediate classes
    """

    def __init__(self, name, expr, dependent=None):
        """
        Create an Intermediate with an associated name

        Arguments
        ---------
        name : str
            The name of the Expression
        expr : sympy.Basic
            The expression
        dependent : ODEObject
            If given the count of this Intermediate will follow as a
            fractional count based on the count of the dependent object
        """
        super(Intermediate, self).__init__(name, expr, dependent)


class StateSolution(Intermediate):
    """
    Sub class of Expression for state solution expressions
    """

    def __init__(self, state, expr, dependent=None):
        """
        Create a StateSolution

        Arguments
        ---------
        state : State
            The state that is being solved for
        expr : sympy.Basic
            The expression that should equal 0 and which solves the state
        dependent : ODEObject
            If given the count of this StateSolution will follow as a
            fractional count based on the count of the dependent object
        """

        check_arg(state, State, 0, StateSolution)
        super(StateSolution, self).__init__(sympycode(state.sym), expr)

        # Flag solved state
        state._is_solved = True
        self._state = state

    @property
    def state(self):
        return self._state

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "'{0}', {1}".format(repr(self.state), sympycode(self.expr))


class DerivativeExpression(Intermediate):
    """
    A class for Intermediate derivative expressions
    """

    def __init__(self, der_expr, dep_var, expr, dependent=None):
        """
        Create a DerivativeExpression

        Arguments
        ---------
        der_expr : Expression, State
            The Expression or State which is differentiated
        dep_var : State, Time, Expression
            The dependent variable
        expr : sympy.Basic
            The expression which the differetiation should be equal
        dependent : ODEObject
            If given the count of this DerivativeExpression will follow as a
            fractional count based on the count of the dependent object
        """
        check_arg(der_expr, Expression, 0, DerivativeExpression)
        check_arg(dep_var, (State, Expression, Time), 1, DerivativeExpression)

        # Check that the der_expr is dependent on var
        if dep_var.sym not in der_expr.sym.args:
            error(
                "Cannot create a DerivativeExpression as {0} is not "
                "dependent on {1}".format(der_expr, dep_var)
            )

        der_sym = sp.Derivative(der_expr.sym, dep_var.sym)
        self._der_expr = der_expr
        self._dep_var = dep_var

        super(DerivativeExpression, self).__init__(sympycode(der_sym), expr, dependent)
        self._sym = sp.Derivative(der_expr.sym, dep_var.sym)
        self._sym._assumptions["real"] = True
        self._sym._assumptions["commutative"] = True
        self._sym._assumptions["imaginary"] = False
        self._sym._assumptions["hermitian"] = True
        self._sym._assumptions["complex"] = True

    @property
    def der_expr(self):
        return self._der_expr

    @property
    def dep_var(self):
        return self._dep_var

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "{0}, {1}, {2}".format(
            repr(self._der_expr), repr(self._dep_var), sympycode(self.expr)
        )

    def _repr_latex_name(self):
        return "\\frac{{d{0}}}{{d{1}}}".format(
            latex(self._der_expr.name), latex(self._dep_var.name)
        )


class RateExpression(Intermediate):
    """
    A sub class of Expression holding single rates
    """

    def __init__(self, to_state, from_state, expr, dependent=None):

        check_arg(to_state, (State, StateSolution), 0, RateExpression)
        check_arg(from_state, (State, StateSolution), 1, RateExpression)

        super(RateExpression, self).__init__(
            "rates_{0}_{1}".format(to_state, from_state), expr, dependent
        )
        self._to_state = to_state
        self._from_state = from_state

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "{0}, {1}, {2}".format(
            repr(self._to_state), repr(self._from_state), sympycode(self.expr)
        )

    @property
    def states(self):
        """
        Return a tuple of the two states the rate expression describes the rate
        between
        """
        return self._to_state, self._from_state


class StateExpression(Expression):
    """
    An expression which determines a State.
    """

    def __init__(self, name, state, expr, dependent=None):
        """
        Create an StateExpression with an assosiated name

        Arguments
        ---------
        name : str
            The name of the state expression. A symbol based on the name will
            be created and accessible to build other expressions
        state : State
            The state which the expression should determine
        expr : sympy.Basic
            The mathematical expression
        dependent : ODEObject
            If given the count of this StateExpression will follow as a
            fractional count based on the count of the dependent object
        """

        check_arg(state, State, 0, StateExpression)

        super(StateExpression, self).__init__(name, expr, dependent)
        self._state = state

    @property
    def state(self):
        return self._state

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "'{0}', {1}, {2}".format(self.name, repr(state), sympycode(self.expr))


class StateDerivative(StateExpression):
    """
    A class for all state derivatives
    """

    def __init__(self, state, expr, dependent=None):
        """
        Create a StateDerivative

        Arguments
        ---------
        state : State
            The state for which the StateDerivative should apply
        expr : sympy.Basic
            The expression which the differetiation should be equal
        dependent : ODEObject
            If given the count of this StateDerivative will follow as a
            fractional count based on the count of the dependent object
        """

        check_arg(state, State, 0, StateDerivative)
        sym = sp.Derivative(state.sym, state.time.sym)
        sym._assumptions["real"] = True
        sym._assumptions["imaginary"] = False
        sym._assumptions["commutative"] = True
        sym._assumptions["hermitian"] = True
        sym._assumptions["complex"] = True

        # Call base class constructor
        super(StateDerivative, self).__init__(sympycode(sym), state, expr, dependent)
        self._sym = sym

    @property
    def sym(self):
        return self._sym

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "{0}, {1}".format(repr(self._state), sympycode(self.expr))

    def _repr_latex_name(self):
        return "\\frac{{d{0}}}{{dt}}".format(latex(self.state.name))


class AlgebraicExpression(StateExpression):
    """
    A class for algebraic expressions which relates a State with an
    expression which should equal to 0
    """

    def __init__(self, state, expr, dependent=None):
        """
        Create an AlgebraicExpression

        Arguments
        ---------
        state : State
            The State which the algebraic expression should determine
        expr : sympy.Basic
            The expression that should equal 0
        dependent : ODEObject
            If given the count of this StateDerivative will follow as a
            fractional count based on the count of the dependent object
        """
        check_arg(state, State, 0, AlgebraicExpression)

        super(AlgebraicExpression, self).__init__(
            "alg_{0}_0".format(state), state, expr, dependent
        )

        # Check that the expr is dependent on the state
        # FIXME: No need because Simone says so!
        # if state.sym not in self.sym:
        #    error("Cannot create an AlgebraicExpression as {0} is not "\
        #          "dependent on {1}".format(state, expr))

    def _args_str(self):
        """
        Return a formatted str of __init__ arguments
        """
        return "{0}, {1}".format(repr(self._state), sympycode(self.expr))

    def _repr_latex_name(self):
        return "0"


class IndexedExpression(IndexedObject, Expression):
    """
    An expression which represents an expression with a fixed index
    associated with it
    """

    def __init__(
        self,
        basename,
        indices,
        expr,
        shape=None,
        array_params=None,
        add_offset="",
        dependent=None,
        enum=None,
    ):
        """
        Create an IndexedExpression with an associated basename used in code
        generation.

        Arguments
        ---------
        basename : str
            The basename of the multi index Expression
        indices : tuple of ints
            The indices
        expr : sympy.Basic
            The expression
        shape : tuple (optional)
            A tuple with the shape of the indexed expression
        array_params : dict
            Parameters to create the array name for the indexed object
        add_offset : bool, str
            If True a fixed offset is added to the indices
        dependent : ODEObject
            If given the count of this IndexedExpression will follow as a
            fractional count based on the count of the dependent object
        enum : str
            String that can be used for enumeration
        """
        IndexedObject.__init__(
            self, basename, indices, shape, array_params, add_offset, dependent, enum
        )
        Expression.__init__(self, self.name, expr, dependent)


class StateIndexedExpression(StateIndexedObject, Expression):
    """
    An expression which represents an expression with a fixed state index
    associated with it
    """

    def __init__(
        self,
        basename,
        indices,
        expr,
        state,
        shape=None,
        array_params=None,
        add_offset="",
        dependent=None,
    ):
        """
        Create an IndexedExpression with an associated basename used in code
        generation.

        Arguments
        ---------
        basename : str
            The basename of the multi index Expression
        indices : tuple of ints
            The indices
        expr : sympy.Basic
            The expression
        state : State
            The state the expression index corresponds to. Used for enumeration.
        shape : tuple (optional)
            A tuple with the shape of the indexed expression
        array_params : dict
            Parameters to create the array name for the indexed object
        add_offset : bool, str
            If True a fixed offset is added to the indices
        dependent : ODEObject
            If given the count of this IndexedExpression will follow as a
            fractional count based on the count of the dependent object

        """

        StateIndexedObject.__init__(
            self, basename, indices, state, shape, array_params, add_offset, dependent
        )
        Expression.__init__(self, self.name, expr, dependent)


class ParameterIndexedExpression(ParameterIndexedObject, Expression):
    """
    An expression which represents an expression with a fixed state index
    associated with it
    """

    def __init__(
        self,
        basename,
        indices,
        expr,
        parameter,
        shape=None,
        array_params=None,
        add_offset="",
        dependent=None,
    ):
        """
        Create an IndexedExpression with an associated basename used in code
        generation.

        Arguments
        ---------
        basename : str
            The basename of the multi index Expression
        indices : tuple of ints
            The indices
        expr : sympy.Basic
            The expression
        parameter : Parameter
            The parameter the expression index corresponds to. Used for enumeration.
        shape : tuple (optional)
            A tuple with the shape of the indexed expression
        array_params : dict
            Parameters to create the array name for the indexed object
        add_offset : bool, str
            If True a fixed offset is added to the indices
        dependent : ODEObject
            If given the count of this IndexedExpression will follow as a
            fractional count based on the count of the dependent object
        """

        ParameterIndexedObject.__init__(
            self,
            basename,
            indices,
            parameter,
            shape,
            array_params,
            add_offset,
            dependent,
        )
        Expression.__init__(self, self.name, expr, dependent)


# Tuple with Derivative types, for type checking
Derivatives = (StateDerivative, DerivativeExpression)



# Copyright (C) 2008-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["adict", "odict"]


class adict(dict):
    """A dictionary with attribute-style access. It maps attribute access to
    the real dictionary.  """

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)

    def __getstate__(self):
        return list(self.__dict__.items())

    def __setstate__(self, items):
        for key, val in items:
            self.__dict__[key] = val

    def __repr__(self):
        return dict.__repr__(self)

    def __setitem__(self, key, value):
        return super(adict, self).__setitem__(key, value)

    def __getitem__(self, name):
        return super(adict, self).__getitem__(name)

    def __delitem__(self, name):
        return super(adict, self).__delitem__(name)

    __getattr__ = __getitem__
    __setattr__ = __setitem__

    def copy(self):
        ch = adict(self)
        return ch


class odict(dict):
    "A simple ordered dict"

    def __init__(self, iterable=None):
        self._keys = []
        dict.__init__(self)
        if iterable is not None:
            for key, value in iterable:
                self[key] = value

    def __setitem__(self, key, value):
        if not key in self._keys:
            self._keys.append(key)
        dict.__setitem__(self, key, value)

    def keys(self):
        return self._keys

    def __iter__(self):
        return iter(self.keys())

    def iterkeys(self):
        for key in self._keys:
            yield key

    def iteritems(self):
        for key in self._keys:
            yield key, self[key]

    def itervalues(self):
        for key in self._keys:
            yield self[key]

    def values(self):
        return [self[key] for key in self._keys]

    def items(self):
        return [(key, self[key]) for key in self._keys]

    def pop(self, *args, **kwargs):
        return NotImplemented

    def popitem(self, *args, **kwargs):
        return NotImplemented

    def setdefault(self, *args, **kwargs):
        return NotImplemented

    def update(self, *args, **kwargs):
        return NotImplemented

    def clear(self):
        self._keys = []
        dict.clear(self)



# Copyright (C) 2012-2014 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ODE"]

# System imports
from collections import defaultdict
import weakref
from functools import cmp_to_key

from sympy.core.function import AppliedUndef

# ModelParameters imports
from modelparameters.sympytools import sp
from modelparameters.codegeneration import sympycode, pythoncode
from modelparameters.utils import Timer

# Local imports
from gotran.common import error, debug, check_arg, check_kwarg, scalars
from gotran.model.odeobjects import *
from gotran.model.expressions import *
from gotran.model.odecomponent import *


class ODE(ODEComponent):
    """
    Root ODEComponent

    Arguments
    ---------
    name : str
        The name of the ODE
    ns : dict (optional)
        A namespace which will be filled with declared ODE symbols
    """

    def __new__(cls, *args, **kwargs):
        self = object.__new__(cls)

        return self

    def __init__(self, name, ns=None):

        # Call super class with itself as parent component
        super(ODE, self).__init__(name, self)

        # Turn off magic attributes (see __setattr__ method) during
        # construction
        self._allow_magic_attributes = False

        # If namespace provided just keep a weak ref
        if ns is None:
            self._ns = {}
        else:
            self._ns = weakref.ref(ns)

        # Add Time object
        # FIXME: Add information about time unit dimensions and make
        # FIXME: it possible to have different time names
        time = Time("t", "ms")
        self._time = time
        self.ode_objects.append(time)

        dt = Dt("dt", "ms")
        self._dt = dt
        self.ode_objects.append(dt)

        # Add to object to component map
        self.object_component = weakref.WeakValueDictionary()
        self.object_component[self._time] = self
        self.object_component[self._dt] = self

        # Namespace, which can be used to eval an expression
        self.ns.update({"t": time.sym, "time": time.sym, "dt": dt.sym})

        # An list with all component names with expression added to them
        # The components are always sorted wrt last expression added
        self.all_expr_components_ordered = []

        # A dict with all components objects
        self.all_components = weakref.WeakValueDictionary()
        self.all_components[name] = self

        # An attribute keeping track of the present ODE component
        self._present_component = self.name

        # A dict with the present ode objects
        # NOTE: hashed by name so duplicated expressions are not stored
        self.present_ode_objects = dict(t=self._time, time=self._time, dt=self._dt)

        # Keep track of duplicated expressions
        self.duplicated_expressions = defaultdict(list)

        # Keep track of expression dependencies and in what expression
        # an object has been used in
        self.expression_dependencies = defaultdict(set)
        self.object_used_in = defaultdict(set)

        # All expanded expressions
        self._expanded_expressions = dict()

        # Attributes which will be populated later
        self._mass_matrix = None

        # Global finalized flag
        self._is_finalized_ode = False

        # Turn on magic attributes (see __setattr__ method)
        self._allow_magic_attributes = True

    @property
    def ns(self):
        if isinstance(self._ns, dict):
            return self._ns

        # Check if ref in weakref is alive
        ns = self._ns()
        if isinstance(ns, dict):
            return ns

        # If not just return an empty dict
        return {}

    @property
    def present_component(self):
        """
        Return the present component
        """
        return self.all_components[self._present_component]

    def import_ode(self, ode, prefix="", components=None, **arguments):
        """
        Import a Model into the present Model

        Arguments
        ---------
        ode : str, gotran.ODE
            The ode which should be added. If ode is a str an
            ODE stored in that file will be loaded. If it is an ODE it will be
            added directly to the present ODE.
        prefix : str (optional)
            A prefix which all state, parameters and intermediates are
            prefixed with.
        components : list, tuple of str (optional)
            A list of components which will either be extracted or excluded
            from the imported model. If not given the whole ODE will be imported.
        arguments : dict (optional)
            Optional arguments which can control loading of model
        """

        timer = Timer("Import ode")

        components = components or []
        check_arg(ode, (str, ODE), 0, context=ODE.import_ode)
        check_arg(prefix, str, 1, context=ODE.import_ode)
        check_arg(components, list, 2, context=ODE.import_ode, itemtypes=str)

        # If ode is given directly
        if isinstance(ode, str):
            # If not load external ODE
            from gotran.model.loadmodel import load_ode

            ode = load_ode(ode, **arguments)

        # Postfix prefix with "_" if prefix is not ""
        if prefix and prefix[-1] != "_":
            prefix += "_"

        # If extracting only a certain components
        if components:
            ode = ode.extract_components(ode.name, *components)

        # Subs for expressions
        subs = {}
        old_new_map = {ode: self}

        def add_comp_and_children(added, comp):
            "Help function to recursively add components to ode"

            # Add states and parameters
            for obj in comp.ode_objects:

                # Check if obj already excists as a ODE parameter
                old_obj = self.present_ode_objects.get(str(obj))

                if old_obj and self.object_component[old_obj] == self:
                    new_name = obj.name
                else:
                    new_name = prefix + obj.name

                if isinstance(obj, State):
                    subs[obj.sym] = added.add_state(new_name, obj.param)

                elif isinstance(obj, Parameter):

                    # If adding an ODE parameter
                    if comp == ode:

                        # And parameter name already excists in the present ODE
                        if str(obj) in self.present_ode_objects:

                            # Skip it and add the registered symbol for
                            # substitution
                            subs[obj.sym] = self.present_ode_objects[str(obj)].sym

                        else:

                            # If not already present just add unprefixed name
                            subs[obj.sym] = added.add_parameter(obj.name, obj.param)

                    else:

                        subs[obj.sym] = added.add_parameter(new_name, obj.param)

            # Add child components
            for child in list(comp.children.values()):
                added_child = added.add_component(child.name)

                # Get corresponding component in present ODE
                old_new_map[child] = added_child

                add_comp_and_children(added_child, child)

        # Recursively add states and parameters
        add_comp_and_children(self, ode)

        # Iterate over all components to add expressions
        for comp_name in ode.all_expr_components_ordered:

            comp = ode.all_components[comp_name]

            comp_comment = "Expressions for the {0} " "component".format(comp.name)

            # Get corresponding component in new ODE
            added = old_new_map[comp]

            # Iterate over all objects of the component and save only expressions
            # and comments
            for obj in comp.ode_objects:

                # If saving an expression
                if isinstance(obj, Expression):

                    # The new sympy expression
                    new_expr = obj.expr.xreplace(subs)

                    # If the component is a Markov model
                    if comp.rates:

                        # Do not save State derivatives
                        if isinstance(obj, StateDerivative):
                            continue

                        # Save rate expressions slightly different
                        elif isinstance(obj, RateExpression):

                            states = obj.states
                            added._add_single_rate(
                                subs[states[0].sym], subs[states[1].sym], new_expr
                            )
                            continue

                    # If no prefix we just add the expression by using setattr
                    # magic
                    if prefix == "":
                        setattr(added, str(obj), new_expr)
                        subs[obj.sym] = added.ode_objects.get(obj.name).sym

                    elif isinstance(obj, (StateExpression, StateSolution)):

                        state = subs[obj.state.sym]

                        if isinstance(obj, AlgebraicExpression):
                            subs[obj.sym] = added.add_algebraic(state, new_expr)

                        elif isinstance(obj, StateDerivative):
                            subs[obj.sym] = added.add_derivative(
                                state, added.t, new_expr
                            )

                        elif isinstance(obj, StateSolution):
                            subs[obj.sym] = added.add_state_solution(state, new_expr)
                            print(repr(obj), obj.sym)
                        else:
                            error("Should not reach here...")

                    # Derivatives are tricky. Here the der expr and dep var
                    # need to be registered in the ODE already. But they can
                    # be registered with and without prefix, so we need to
                    # check that.
                    elif isinstance(obj, Derivatives):

                        # Get der_expr
                        der_expr = self.root.present_ode_objects.get(obj.der_expr.name)

                        if der_expr is None:

                            # If not found try prefixed version
                            der_expr = self.root.present_ode_objects.get(
                                prefix + obj.der_expr.name
                            )

                            if der_expr is None:
                                if prefix:
                                    error(
                                        "Could not find expression: "
                                        "({0}){1} while adding "
                                        "derivative".format(prefix, obj.der_expr)
                                    )
                                else:
                                    error(
                                        "Could not find expression: "
                                        "{1} while adding derivative".format(
                                            obj.der_expr
                                        )
                                    )

                        dep_var = self.root.present_ode_objects.get(obj.dep_var.name)

                        if isinstance(dep_var, Time):
                            dep_var = self._time

                        elif dep_var is None:

                            # If not found try prefixed version
                            dep_var = self.root.present_ode_objects.get(
                                prefix + obj.dep_var.name
                            )

                            if dep_var is None:
                                if prefix:
                                    error(
                                        "Could not find expression: "
                                        "({0}){1} while adding "
                                        "derivative".format(prefix, obj.dep_var)
                                    )
                                else:
                                    error(
                                        "Could not find expression: "
                                        "{1} while adding derivative".format(
                                            obj.dep_var
                                        )
                                    )

                        subs[obj.sym] = added.add_derivative(
                            der_expr, dep_var, new_expr
                        )

                    elif isinstance(obj, Intermediate):

                        subs[obj.sym] = added.add_intermediate(
                            prefix + obj.name, new_expr
                        )

                    else:
                        error("Should not reach here!")

                # If saving a comment
                elif isinstance(obj, Comment) and str(obj) != comp_comment:
                    added.add_comment(str(obj))

    def save(self, basename=None):
        """
        Save ODE to file

        Arguments
        ---------
        basename : str (optional)
            The basename of the file which the ode will be saved to, if not
            given the basename will be the same as the name of the ode.
        """

        timer = Timer("Save " + self.name)

        if not self._is_finalized_ode:
            error("ODE need to be finalized to be saved to file.")

        lines = ["# Saved Gotran model"]

        comp_names = dict()

        basename = basename or self.name

        for comp in self.components:
            if comp == self:
                comp_name = ""
            else:
                present_comp = comp
                comps = [present_comp.name]
                while present_comp.parent != self:
                    present_comp = present_comp.parent
                    comps.append(present_comp.name)
                comp_name = ", ".join('"{0}"'.format(name) for name in reversed(comps))

            comp_names[comp] = comp_name

            states = [
                "{0}={1},".format(obj.name, obj.param.repr(include_name=False))
                for obj in comp.ode_objects
                if isinstance(obj, State)
            ]
            parameters = [
                "{0}={1},".format(obj.name, obj.param.repr(include_name=False))
                for obj in comp.ode_objects
                if isinstance(obj, Parameter)
            ]
            if states:
                lines.append("")
                if comp_name:
                    lines.append("states({0},".format(comp_name))
                else:
                    lines.append("states({0}".format(states.pop(0)))
                for state_code in states:
                    lines.append("       " + state_code)
                lines[-1] = lines[-1][:-1] + ")"

            if parameters:
                lines.append("")
                if comp_name:
                    lines.append("parameters({0},".format(comp_name))
                else:
                    lines.append("parameters({0}".format(parameters.pop(0)))
                for param_code in parameters:
                    lines.append("           " + param_code)
                lines[-1] = lines[-1][:-1] + ")"

        # Iterate over all components
        for comp_name in self.all_expr_components_ordered:

            comp = self.all_components[comp_name]

            comp_comment = "Expressions for the {0} " "component".format(comp.name)

            # Iterate over all objects of the component and save only expressions
            # and comments
            for obj in comp.ode_objects:

                # If saving an expression
                if isinstance(obj, Expression):

                    # If the component is a Markov model
                    if comp.rates:

                        # Do not save State derivatives
                        if isinstance(obj, StateDerivative):
                            continue

                        # Save rate expressions slightly different
                        elif isinstance(obj, RateExpression):
                            lines.append(
                                "rates[{0}, {1}] = {2}".format(
                                    sympycode(obj.states[0]),
                                    sympycode(obj.states[1]),
                                    sympycode(obj.expr),
                                )
                            )
                            continue

                    # All other Expressions
                    lines.append("{0} = {1}".format(obj.name, sympycode(obj.expr)))

                # If saving a comment
                elif isinstance(obj, Comment):

                    # If comment is component comment
                    if str(obj) == comp_comment:
                        lines.append("")
                        comp_name = (
                            comp_names[comp]
                            if comp_names[comp]
                            else '"{0}"'.format(basename)
                        )
                        lines.append("expressions({0})".format(comp_name))

                    # Just add the comment
                    else:
                        lines.append("")
                        lines.append('comment("{0}")'.format(obj))

        lines.append("")
        # Use Python code generator to indent outputted code
        # Write to file
        from gotran.codegeneration.codegenerators import PythonCodeGenerator

        with open(basename + ".ode", "w") as f:
            f.write("\n".join(PythonCodeGenerator.indent_and_split_lines(lines)))

    def register_ode_object(self, obj, comp, dependent=None):
        """
        Register an ODE object in the root ODEComponent
        """

        from modelparameters.sympytools import symbols_from_expr

        if self._is_finalized_ode and isinstance(obj, StateExpression):
            error("Cannot register a StateExpression, the ODE is finalized")

        # Check for existing object in the ODE
        dup_obj = self.present_ode_objects.get(obj.name)

        # If object with same name is already registered in the ode we
        # need to figure out what to do
        if dup_obj:

            try:
                dup_comp = self.object_component[dup_obj]
            except KeyError:
                dup_comp = None

            # If a state is substituted by a state solution
            if isinstance(dup_obj, State) and isinstance(obj, StateSolution):
                debug("Reduce state '{0}' to {1}".format(dup_obj, obj.expr))

            # If duplicated object is an ODE Parameter and the added object is
            # either a State or a Parameter we replace the Parameter.
            elif (
                isinstance(dup_obj, Parameter)
                and dup_comp == self
                and comp != self
                and isinstance(obj, (State, Parameter))
            ):

                timer = Timer("Replace objects")

                # Remove the object
                self.ode_objects.remove(dup_obj)

                # FIXME: Do we need to recreate all expression the objects is used in?
                # Replace the object from the object_used_in dict and update
                # the correponding expressions
                subs = {dup_obj.sym: obj.sym}
                subs = {}

                # Recursively replace object dependencies
                self._replace_object(dup_obj, obj, subs)

                # for expr in self.object_used_in[dup_obj]:
                #    updated_expr = recreate_expression(expr, subs)
                #    self.object_used_in[obj].add(updated_expr)
                #
                #    # Exchange and update the dependencies
                #    self.expression_dependencies[expr].remove(dup_obj)
                #    self.expression_dependencies[expr].add(obj)
                #
                #    # FIXME: Do not remove the dependencies
                #    #self.expression_dependencies[updated_expr] = \
                #    #            self.expression_dependencies.pop(expr)
                #    self.expression_dependencies[updated_expr] = \
                #                self.expression_dependencies[expr]
                #
                #    # Find the index of old expression and exchange it with updated
                #    old_comp = self.object_component[expr]
                #    ind = old_comp.ode_objects.index(expr)
                #    old_comp.ode_objects[ind] = updated_expr
                #
                ## Remove information about the replaced objects
                # self.object_used_in.pop(dup_obj)

            # If duplicated object is an ODE Parameter and the added
            # object is an Intermediate we raise an error.
            elif (
                isinstance(dup_obj, Parameter)
                and dup_comp == self
                and isinstance(obj, Expression)
            ):
                error(
                    "Cannot replace an ODE parameter with an Expression, "
                    "only with Parameters and States."
                )

            # If State, Parameter or DerivativeExpression we always raise an error
            elif any(
                isinstance(
                    oo,
                    (
                        State,
                        Parameter,
                        Time,
                        Dt,
                        DerivativeExpression,
                        AlgebraicExpression,
                        StateSolution,
                    ),
                )
                for oo in [dup_obj, obj]
            ):
                error(
                    "Cannot register {0}. A {1} with name '{2}' is "
                    "already registered in this ODE.".format(
                        type(obj).__name__, type(dup_obj).__name__, dup_obj.name
                    )
                )
            else:

                # Sanity check that both obj and dup_obj are Expressions
                assert all(isinstance(oo, (Expression)) for oo in [dup_obj, obj])

                # Get list of duplicated objects or an empy list
                dup_objects = self.duplicated_expressions[obj.name]
                if len(dup_objects) == 0:
                    dup_objects.append(dup_obj)
                dup_objects.append(obj)

        # Update global information about ode object
        self.present_ode_objects[obj.name] = obj
        self.object_component[obj] = comp
        self.ns.update({obj.name: obj.sym})

        # If Expression
        if isinstance(obj, Expression):

            # Append the name to the list of all ordered components with
            # expressions. If the ODE is finalized we do not update components
            if not self._is_finalized_ode:
                self._handle_expr_component(comp, obj)

            # Expand and add any derivatives in the expressions
            expression_added = False
            replace_dict = {}
            derivative_expression_list = list(obj.expr.atoms(sp.Derivative))
            derivative_expression_list.sort(key=lambda e: e.sort_key())
            for der_expr in derivative_expression_list:
                expression_added |= self._expand_single_derivative(
                    comp, obj, der_expr, replace_dict, dependent
                )

            # If expressions need to be re-created
            if replace_dict:
                obj.replace_expr(replace_dict)

            # If any expression was added we need to bump the count of the ODEObject
            if expression_added:
                obj._recount(dependent=dependent)

            # Add dependencies between the last registered comment and
            # expressions so they are carried over in Code components
            if comp._local_comments:
                self.object_used_in[comp._local_comments[-1]].add(obj)
                self.expression_dependencies[obj].add(comp._local_comments[-1])

            # Get expression dependencies
            for sym in symbols_from_expr(obj.expr, include_derivatives=True):

                dep_obj = self.present_ode_objects[sympycode(sym)]

                if dep_obj is None:
                    error(
                        "The symbol '{0}' is not declared within the '{1}' "
                        "ODE.".format(sym, self.name)
                    )

                # Store object dependencies
                self.expression_dependencies[obj].add(dep_obj)
                self.object_used_in[dep_obj].add(obj)

            # If the expression is a StateSolution the state cannot have
            # been used previously
            if isinstance(obj, StateSolution) and self.object_used_in.get(obj.state):
                used_in = self.object_used_in.get(obj.state)
                error(
                    "A state solution cannot have been used in "
                    "any previous expressions. {0} is used in: {1}".format(
                        obj.state, used_in
                    )
                )

    def expanded_expression(self, expr):
        """
        Return the expanded expression.

        The returned expanded expression consists of the original
        expression given by it basics objects (States, Parameters and
        IndexedObjects)
        """

        timer = Timer("Expand expression")

        check_arg(expr, Expression)

        # First check cache
        exp_expr = self._expanded_expressions.get(expr)

        if exp_expr is not None:
            return exp_expr

        # If not recursively expand the expression
        der_subs = {}
        subs = {}
        for obj in self.expression_dependencies[expr]:

            if isinstance(obj, Derivatives):
                der_subs[obj.sym] = self.expanded_expression(obj)

            elif isinstance(obj, Expression):
                subs[obj.sym] = self.expanded_expression(obj)

        # Do the substitution
        exp_expr = expr.expr.xreplace(der_subs).xreplace(subs)
        self._expanded_expressions[expr] = exp_expr

        return exp_expr

    def extract_components(self, name, *components):
        """
        Create an ODE from a number of components

        Returns an ODE including the components

        Arguments
        ---------
        name : str
            The name of the created ODE
        components : str
            A variable len tuple of str describing the components
        """
        check_arg(name, str, 0)
        check_arg(components, tuple, 1, itemtypes=str)

        components = list(components)

        collected_components = set()

        if self.name in components:
            error("Can only extract sub component of this ODE.")

        # Collect components and check that the ODE has the components
        for original_component in self.components:

            # Collect components together with its children
            if original_component.name in components:

                components.remove(original_component.name)
                collected_components.update(original_component.components)

        # Check that there are no components left
        if components:
            if len(components) > 1:
                error(
                    "{0} are not a components of this ODE.".format(
                        ", ".join("'{0}'".format(comp) for comp in components)
                    )
                )
            else:
                error("'{0}' is not a component of this ODE.".format(components[0]))

        # Collect dependencies
        included_objects = []
        dependencies = set()
        for comp in self.components:
            if comp in collected_components:
                included_objects.extend(comp.ode_objects)
                for expr in comp.ode_objects:
                    if isinstance(expr, Expression):
                        dependencies.update(self.expression_dependencies[expr])

        # Remove included objects from dependencies
        dependencies.difference_update(included_objects)

        # Create return ODE
        ode = ODE(name)

        # Add dependencies as parameters to return ODE
        subs = dict()
        for dep in dependencies:

            # Skip time
            if str(dep) in ["t", "time", "dt"]:
                continue
            subs[dep.sym] = ode.add_parameter(dep.name, dep.param.value)

        # Add components together with states and parameters to the ODE
        components = sorted(collected_components, reverse=True)
        old_new_map = dict()

        def add_comp_and_children(comp, components, parent):
            "Help function to add recursively components to ode"

            # Add component
            added = parent.add_component(comp.name)
            old_new_map[comp] = added

            # Add states and parameters
            for obj in comp.ode_objects:
                if isinstance(obj, State):
                    added.add_state(obj.name, obj.param)
                elif isinstance(obj, Parameter):
                    added.add_parameter(obj.name, obj.param)

            # Remove the added component
            components.remove(comp)

            for child in list(comp.children.values()):
                add_comp_and_children(child, components, added)

        # Add component recursivly
        while components:
            add_comp_and_children(components[-1], components, ode)

        # Iterate over all components to add expressions
        for comp_name in self.all_expr_components_ordered:

            comp = self.all_components[comp_name]

            # If we should add component
            if comp not in collected_components:
                continue

            comp_comment = "Expressions for the {0} " "component".format(comp.name)

            # Get corresponding component in new ODE
            added = old_new_map[comp]

            # Iterate over all objects of the component and save only
            # expressions and comments
            for obj in comp.ode_objects:

                # If saving an expression
                if isinstance(obj, Expression):

                    # The new sympy expression
                    new_expr = obj.expr.xreplace(subs)

                    # If the component is a Markov model
                    if comp.rates:

                        # Do not save State derivatives
                        if isinstance(obj, StateDerivative):
                            continue

                        # Save rate expressions slightly different
                        elif isinstance(obj, RateExpression):

                            states = obj.states
                            added._add_single_rate(
                                states[0].sym, states[1].sym, new_expr
                            )
                            continue

                    # All other Expressions
                    setattr(added, str(obj), new_expr)

                # If saving a comment
                elif isinstance(obj, Comment) and str(obj) != comp_comment:
                    added.add_comment(str(obj))

        # Finalize ode and return it
        ode.finalize()
        return ode

    @property
    def mass_matrix(self):
        """
        Return the mass matrix as a sympy.Matrix
        """

        if not self.is_finalized:
            error("The ODE must be finalized")

        if not self._mass_matrix:

            state_exprs = self.state_expressions
            N = len(state_exprs)
            self._mass_matrix = sp.Matrix(
                N,
                N,
                lambda i, j: 1
                if i == j and isinstance(state_exprs[i], StateDerivative)
                else 0,
            )

        return self._mass_matrix

    @property
    def is_dae(self):
        """
        Return True if ODE is a DAE
        """
        if not self.is_complete:
            error("The ODE is not complete")

        return any(
            isinstance(expr, AlgebraicExpression) for expr in self.state_expressions
        )

    def finalize(self):
        """
        Finalize the ODE
        """
        for comp in self.components:
            comp.finalize_component()

        self._is_finalized_ode = True
        self._present_component = self.name

    def signature(self):
        """
        Return a signature uniquely defining the ODE
        """
        import hashlib

        def_list = []
        for comp in self.components:
            if self != comp:
                def_list.append(str(comp))

            # Sort wrt stringified states and parameters avoiding trouble with
            # random ordering of **kwargs
            def_list += sorted([repr(state.param) for state in comp.full_states])
            def_list += sorted([repr(param.param) for param in comp.parameters])
            def_list += [sympycode(expr.expr) for expr in comp.intermediates]

            # Sort state expressions wrt stringified state names
            def_list += [
                sympycode(expr.expr)
                for expr in sorted(
                    self.state_expressions,
                    key=cmp_to_key(lambda o0, o1: cmp(str(o0.state), str(o1.state))),
                )
            ]

        h = hashlib.sha1()
        h.update(";".join(def_list).encode("utf-8"))
        return h.hexdigest()

    def _replace_object(self, old_obj, replaced_obj, replace_dicts):
        """
        Recursivley replace an expression by recreating the expression
        using passed replace dicts
        """

        # Iterate over expressions obj is used in and replace these
        self.present_ode_objects[old_obj.name] = replaced_obj
        replace_dicts[old_obj.sym] = replaced_obj.sym
        self.object_component[replaced_obj] = self.object_component.pop(old_obj)

        for old_expr in self.object_used_in[old_obj]:

            # Recreate expression
            replaced_expr = recreate_expression(old_expr, replace_dicts)

            # Update used_in dict
            self.object_used_in[replaced_obj].add(replaced_expr)

            # Update all object used in
            for dep in self.expression_dependencies[old_expr]:
                if isinstance(dep, Comment):
                    continue
                if old_expr in self.object_used_in[dep]:
                    self.object_used_in[dep].remove(old_expr)
                    self.object_used_in[dep].add(replaced_expr)

            # Exchange and update the dependencies
            if old_obj in self.expression_dependencies[old_expr]:
                self.expression_dependencies[old_expr].remove(old_obj)
            self.expression_dependencies[old_expr].add(replaced_obj)

            # FIXME: Do not remove the dependencies
            # self.expression_dependencies[updated_expr] = \
            #            self.expression_dependencies.pop(expr)
            self.expression_dependencies[replaced_expr] = self.expression_dependencies[
                old_expr
            ]

            # Find the index of old expression and exchange it with updated
            old_comp = self.object_component[old_expr]

            # Get indexed of old expr and overwrite it with new expr
            ind = old_comp.ode_objects.index(old_expr)
            old_comp.ode_objects[ind] = replaced_expr

            # Update the expressions this expression is used in
            self._replace_object(old_expr, replaced_expr, replace_dicts)

        # Remove information about the replaced objects
        self.object_used_in.pop(old_obj)

    def _handle_expr_component(self, comp, expr):
        """
        A help function to sort and add components in the ordered
        the intermediate expressions are added to the ODE
        """

        if len(self.all_expr_components_ordered) == 0:
            self.all_expr_components_ordered.append(comp.name)

            # Add a comment to the component
            comp.add_comment("Expressions for the {0} " "component".format(comp.name))

            # Recount the last added expression so the comment comes
            # infront of the expression
            expr._recount()

        # We are shifting expression components
        elif self.all_expr_components_ordered[-1] != comp.name:

            # Finalize the last component we visited
            self.all_components[
                self.all_expr_components_ordered[-1]
            ].finalize_component()

            # Append this component
            self.all_expr_components_ordered.append(comp.name)

            # Add a comment to the component
            comp.add_comment("Expressions for the {0} " "component".format(comp.name))

            # Recount the last added expression so the comment comes
            # infront of the expression
            expr._recount()

    def _expand_single_derivative(self, comp, obj, der_expr, replace_dict, dependent):
        """
        Expand a single derivative and register it as new derivative expression

        Returns True if an expression was actually added

        Populate replace dict with a replacement for the derivative if it is trivial
        """

        # Try accessing already registered derivative expressions
        der_expr_obj = self.present_ode_objects.get(sympycode(der_expr))

        # If excist continue
        if der_expr_obj:
            return False

        # Try expand the given derivative if it is directly expandable just
        # add a replacement for the derivative result
        der_result = der_expr.args[0].diff(der_expr.args[1])
        if not der_result.atoms(sp.Derivative):
            replace_dict[der_expr] = der_result
            return False

        if not isinstance(der_expr.args[0], AppliedUndef):
            error(
                "Can only register Derivatives of allready registered "
                "Expressions. Got: {0}".format(sympycode(der_expr.args[0]))
            )

        if not isinstance(der_expr.args[1], (AppliedUndef, sp.Symbol)):
            error(
                "Can only register Derivatives with a single dependent "
                "variabe. Got: {0}".format(sympycode(der_expr.args[1]))
            )

        # Get the expr and dependent variable objects
        expr_obj = self.present_ode_objects[sympycode(der_expr.args[0])]
        var_obj = self.present_ode_objects[sympycode(der_expr.args[1])]

        # If the dependent variable is time and the expression is a state
        # variable we raise an error as the user should already have created
        # the expression.
        if isinstance(expr_obj, State) and var_obj == self._time:
            error(
                "The expression {0} is dependent on the state "
                "derivative of {1} which is not registered in this ODE.".format(
                    obj, expr_obj
                )
            )

        # If we get a Derivative(expr, t) we issue an error
        # if isinstance(expr_obj, Expression) and var_obj == self._time:
        #    error("All derivative expressions of registered expressions "\
        #          "need to be expanded with respect to time. Use "\
        #          "expr.diff(t) instead of Derivative(expr, t) ")

        if not isinstance(expr_obj, Expression):
            error(
                "Can only differentiate expressions or states. Got {0} as "
                "the derivative expression.".format(expr_obj)
            )

        # Expand derivative and see if it is trivial
        der_result = expr_obj.expr.diff(var_obj.sym)

        # If derivative result are trival we substitute it
        if (
            der_result.is_number
            or isinstance(der_result, (sp.Symbol, AppliedUndef))
            or (
                isinstance(der_result, (sp.Mul, sp.Pow, sp.Add))
                and len(der_result.args) == 2
                and all(
                    isinstance(arg, (sp.Number, sp.Symbol, AppliedUndef))
                    for arg in der_result.args
                )
            )
        ):
            replace_dict[der_expr] = der_result
            return False

        # Store expression
        comp.add_derivative(expr_obj, var_obj, der_result, dependent)

        return True



# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["load_ode", "exec_ode", "load_cell"]

# System imports
import inspect
import os
import shutil
import re
import tempfile
import weakref
from collections import OrderedDict, deque

# modelparameters import
from modelparameters.parameters import (
    Param,
    ScalarParam,
    ArrayParam,
    ConstParam,
    scalars,
    OptionParam,
)
from modelparameters.sympytools import sp_namespace, sp

# gotran imports
from gotran.common import *
from gotran.model.ode import ODE
from gotran.model.cellmodel import CellModel

_for_template = re.compile("\A[ ]*for[ ]+.*in[ ]+:[ ]*\Z")
_no_intermediate_template = re.compile(".*# NO INTERMEDIATE.*")


class IntermediateDispatcher(dict):
    """
    Dispatch intermediates to ODE attributes
    """

    def __init__(self, ode=None):
        """
        Initalize with an ode
        """
        if ode is not None:
            self._ode = weakref.ref(ode)
        else:
            self._ode = None

    @property
    def ode(self):

        if self._ode == None:
            error("ode attr is not set")

        return self._ode()

    @ode.setter
    def ode(self, ode):
        self._ode = weakref.ref(ode)

    def __setitem__(self, name, value):
        """
        This is only for expressions
        """

        from modelparameters.sympytools import symbols_from_expr

        timer = Timer("Namespace dispatcher")
        # Set the attr of the ODE
        # If a scalar or a sympy number or it is a sympy.Basic consisting of
        # sp.Symbols

        if (
            isinstance(value, scalars)
            or isinstance(value, sp.Number)
            or (isinstance(value, sp.Basic) and symbols_from_expr(value))
        ):

            # Get source which triggers the insertion to the global namespace
            frame = inspect.currentframe().f_back
            lines, lnum = inspect.findsource(frame)
            # Try getting the code
            try:
                code = lines[frame.f_lineno - 1].strip()

                # Concatenate lines with line continuation symbols
                prev = 2
                while (
                    frame.f_lineno - prev >= 0
                    and len(lines[frame.f_lineno - prev]) >= 2
                    and lines[frame.f_lineno - prev][-2:] == "\\\n"
                ):
                    code = lines[frame.f_lineno - prev][:-2].strip() + code
                    prev += 1
            except:
                code = ""

            # Check if the line includes a for statement
            # Here we strip op potiential code comments after the main for
            # statement.
            if re.search(_for_template, code.split("#")[0].strip()) or re.search(
                _no_intermediate_template, code
            ):

                debug("Not registering '{0}' as an intermediate.".format(name))

                # If so just add the value to the namespace without
                # registering the intermediate
                dict.__setitem__(self, name, value)

            else:

                del timer

                # Add obj to the present component
                sym = setattr(self.ode.present_component, name, value)

        else:
            debug("Not registering '{0}' as an intermediate.".format(name))

            # If no ode attr was generated we just add the value to the
            # namespace
            dict.__setitem__(self, name, value)

    def update(self, other):
        check_arg(other, dict)
        for name, value in list(other.items()):
            dict.__setitem__(self, name, value)


def _init_namespace(ode, load_arguments, namespace):
    """
    Create namespace and populate it
    """

    namespace.update(sp_namespace)

    # Add Sympy matrix related stuff
    namespace.update(dict(eye=sp.eye, diag=sp.diag, Matrix=sp.Matrix, zeros=sp.zeros))

    namespace.update(
        dict(
            t=ode.t,
            ScalarParam=ScalarParam,
            ArrayParam=ArrayParam,
            ConstParam=ConstParam,
            OptionParam=OptionParam,
            sp=sp,
        )
    )

    # Add ode and model_arguments
    _namespace_binder(namespace, weakref.ref(ode), load_arguments)

    return namespace


def exec_ode(ode_str, name, **arguments):
    """
    Execute an ode given by a str

    Arguments
    ---------
    ode_str : str
        The ode as a str
    name : str
        The name of ODE
    arguments : dict (optional)
        Optional arguments which can control loading of model
    """
    # Dict to collect declared intermediates
    intermediate_dispatcher = IntermediateDispatcher()

    # Create an ODE which will be populated with data when ode file is loaded
    ode = ODE(name, intermediate_dispatcher)

    intermediate_dispatcher.ode = ode

    debug("Loading {}".format(ode.name))

    # Create namespace which the ode file will be executed in
    _init_namespace(ode, arguments, intermediate_dispatcher)

    # Write str to file
    with open("_tmp_gotrand.ode", "w") as f:
        f.write(ode_str)

    # Execute file and collect
    with open("_tmp_gotrand.ode") as f:
        exec(compile(f.read(), "_tmp_gotrand.ode", "exec"), intermediate_dispatcher)
    os.unlink("_tmp_gotrand.ode")

    # Finalize ODE
    ode.finalize()

    # Check for completeness
    if not ode.is_complete:
        warning("ODE mode '{0}' is not complete.".format(ode.name))

    info("Loaded ODE model '{0}' with:".format(ode.name))
    for what in ["full_states", "parameters"]:
        num = getattr(ode, "num_{0}".format(what))
        info("{0}: {1}".format(("Num " + what.replace("_", " ")).rjust(20), num))
    return ode


def load_ode(filename, name=None, **arguments):
    """
    Load an ODE from file and return the instance

    The method looks for a file with .ode extension.

    Arguments
    ---------
    filename : str
        Name of the ode file to load
    name : str (optional)
        Set the name of ODE (defaults to filename)
    arguments : dict (optional)
        Optional arguments which can control loading of model
    """

    arguments["class_type"] = "ode"
    return _load(filename, name, **arguments)


def _load(filename, name, **arguments):
    """
    Load an ODE or Cell from file and return the instance

    The method looks for a file with .ode extension.

    Arguments
    ---------
    filename : str
        Name of the ode file to load
    name : str (optional)
        Set the name of ODE (defaults to filename)
    arguments : dict (optional)
        Optional arguments which can control loading of model
    """
    timer = Timer("Load ODE")
    # Extract name from filename
    if len(filename) < 4 or filename[-4:] != ".ode":
        name = name or filename
        filename = filename + ".ode"
    elif name is None:
        name = filename[:-4]

    # Execute the file
    if not os.path.isfile(filename):
        error("Could not find '{0}'".format(filename))

    # Copy file temporary to current directory
    basename = os.path.basename(filename)
    copyfile = False
    if not basename == filename:
        shutil.copy(filename, basename)
        filename = basename
        name = filename[:-4]
        copyfile = True

    # If a Param is provided turn it into its value
    for key, value in list(arguments.items()):
        if isinstance(value, Param):
            arguments[key] = value.getvalue()

    class_type = arguments.pop("class_type", "ode")
    msg = "Argument class_type must be one of " "('ode', 'cell'), got %s " % class_type
    assert class_type in ("ode", "cell"), msg

    # Dict to collect namespace
    intermediate_dispatcher = IntermediateDispatcher()

    # Create an ODE which will be populated with data when ode file is loaded

    if class_type == "ode":
        ode = ODE(name, intermediate_dispatcher)
    else:
        ode = CellModel(name, intermediate_dispatcher)

    intermediate_dispatcher.ode = ode

    debug("Loading {}".format(ode.name))

    # Create namespace which the ode file will be executed in
    _init_namespace(ode, arguments, intermediate_dispatcher)

    # Execute file and collect
    with open(filename, "r") as f:
        exec(compile(f.read(), filename, "exec"), intermediate_dispatcher)

    # Finalize ODE
    ode.finalize()

    # Check for completeness
    if not ode.is_complete:
        warning("ODE model '{0}' is not complete.".format(ode.name))

    info("Loaded ODE model '{0}' with:".format(ode.name))
    for what in ["full_states", "parameters"]:
        num = getattr(ode, "num_{0}".format(what))
        info("{0}: {1}".format(("Num " + what.replace("_", " ")).rjust(20), num))
    if copyfile:
        os.unlink(filename)

    return ode


def load_cell(filename, name=None, **arguments):
    """
    Load a Cell from file and return the instance

    The method looks for a file with .ode extension.

    Arguments
    ---------
    filename : str
        Name of the ode file to load
    name : str (optional)
        Set the name of ODE (defaults to filename)
    arguments : dict (optional)
        Optional arguments which can control loading of model
    """
    arguments["class_type"] = "cell"

    cell = _load(filename, name, **arguments)
    cell._initialize_cell_model()
    return cell


def _namespace_binder(namespace, ode, load_arguments):
    """
    Add functions all bound to current ode, namespace and arguments
    """

    def comment(comment):
        """
        Add a comment to the present ODE component

        Arguments
        ---------
        comment : str
            The comment
        """

        comp = ode().present_component

        # Add the comment
        comp.add_comment(comment)

    def subode(subode, prefix="", components=None):
        """
        Load an ODE and add it to the present ODE (deprecated)

        Argument
        --------
        subode : str
            The subode which should be added.
        prefix : str (optional)
            A prefix which all state and parameters are prefixed with.
        components : list, tuple of str (optional)
        A list of components which will be extracted and added to the present
            ODE. If not given the whole ODE will be added.
        """
        warning("Usage of 'sub_ode()' is deprecated. " "Use 'import_ode()' instead.")
        import_model(subode, prefix, components)

    def import_ode(subode, prefix="", components=None, **arguments):
        """
        Import an ODE into the present ode

        Argument
        --------
        subode : str
            The ode which should be imported
        prefix : str (optional)
            A prefix which all state, parameters and intermediates are
            prefixed with.
        components : list, tuple of str (optional)
            A list of components which will either be extracted or excluded
            from the imported ode. If not given the whole ODE will be imported.
        arguments : dict (optional)
            Optional arguments which can control loading of model
        """

        check_arg(subode, str, 0)

        # Add the subode and update namespace
        ode().import_ode(subode, prefix=prefix, components=components, **arguments)

    def timeunit(*args, **kwargs):
        """
        Update timeunit according to timeunit in file
        """
        comp = ode()

        if args and isinstance(args[0], str):
            comp._time.update_unit(args[0])
            comp._dt.update_unit(args[0])

    def states(*args, **kwargs):
        """
        Add a number of states to the current component or to the
        chosed component
        """

        # If comp string is passed we get the component
        if args and isinstance(args[0], str):
            comp = ode()
            args = deque(args)

            while args and isinstance(args[0], str):
                comp = comp(args.popleft())
        else:
            comp = ode().present_component

        # Update the rates name so it points to the present components
        # rates dictionary
        namespace["rates"] = comp.rates

        # Add the states
        comp.add_states(*args, **kwargs)

    def parameters(*args, **kwargs):
        """
        Add a number of parameters to the current ODE or to the chosed component
        """
        # If comp string is passed we get the component
        if args and isinstance(args[0], str):
            comp = ode()
            args = deque(args)

            while args and isinstance(args[0], str):
                comp = comp(args.popleft())
        else:
            comp = ode().present_component

        # Update the rates name so it points to the present components
        # rates dictionary
        namespace["rates"] = comp.rates

        # Add the parameters and update namespace
        comp.add_parameters(*args, **kwargs)

    def model_arguments(**kwargs):
        """
        Defines arguments that can be altered while the ODE is loaded

        Example
        -------

        In gotran model file:

          >>> ...
          >>> model_arguments(include_Na=True)
          >>> if include_Na:
          >>>     states(Na=1.0)
          >>> ...

        When the model gets loaded

          >>> ...
          >>> load_ode("model", include_Na=False)
          >>> ...

        """

        # Check the passed load arguments
        for key in load_arguments:
            if key not in kwargs:
                error("Name '{0}' is not a model_argument.".format(key))

        # Update the namespace
        ns = {}
        for key, value in list(kwargs.items()):
            if not isinstance(value, (float, int, str, Param)):
                error(
                    "expected only 'float', 'int', 'str' or 'Param', as model_arguments, "
                    "got: '{}' for '{}'".format(type(value).__name__, key)
                )

            if key not in load_arguments:
                ns[key] = value.getvalue() if isinstance(value, Param) else value

            else:
                # Try to cast the passed load_arguments with the orginal type
                if isinstance(value, Param):

                    # Cast value
                    new_value = value.value_type(load_arguments[key])

                    # Try to set new value
                    value.setvalue(new_value)

                    # Assign actual value of Param
                    ns[key] = value.getvalue()

                else:
                    ns[key] = type(value)(load_arguments[key])

        namespace.update(ns)

    def component(*args):
        """
        Set the present component, deprecated
        """
        warning("Usage of 'component()' is deprecated. " "Use 'expressions()' instead.")
        return expressions(*args)

    def expressions(*args):

        check_arg(args, tuple, 0, itemtypes=str)

        comp = ode()

        args = deque(args)

        while args:
            comp = comp(args.popleft())

        assert ode().present_component == comp

        # Update the rates name so it points to the present components
        # rates dictionary
        namespace["rates"] = comp.rates

        return comp

    # Update provided namespace
    namespace.update(
        dict(
            timeunit=timeunit,
            states=states,
            parameters=parameters,
            model_arguments=model_arguments,
            component=component,
            expressions=expressions,
            subode=subode,
            import_ode=import_ode,
            comment=comment,
        )
    )



# Global imports
import signal
import numpy as np


gotran_methods = [
    "explicit_euler",
    "rush_larsen",
    "generalized_rush_larsen",
    "simplified_implicit_euler",
]

sundials_methods = ["cvode"]

goss_methods = ["RKF32"]

methods = ["scipy"] + gotran_methods + sundials_methods + goss_methods


class ODESolverError(Exception):
    pass


def timeout(timeout_seconds):
    def decorate(function):
        message = "Timeout (%s sec) elapsed for solve %s" % (
            timeout_seconds,
            function.__name__,
        )

        def new_f(*args, **kwargs):
            old = signal.signal(signal.SIGALRM, handler)

            try:
                function_result = function(*args, **kwargs)
            finally:
                signal.signal(signal.SIGALRM, old)
            signal.alarm(0)
            return function_result

        new_f.func_name = function.func_name
        return new_f

    return decorate


class Solver(object):
    def __init__(self, ode, **options):

        # Set 1000 seconds as max
        self.max_solve_time = options.get("max_solve_time", 1000)

        # Get names of monitoed expression
        self._monitored = []
        for expr in sorted(ode.intermediates + ode.state_expressions):
            self._monitored.append(expr.name)

        # Get names of states
        self._state_names = ode.state_symbols

        # Create C module
        self._module = generate_module(ode, self._monitored, **options)

        # The right hand side
        self._rhs = self._module.rhs
        # The jacobian
        self._jac = (
            None
            if not hasattr(self.module, "compute_jacobian")
            else self.module.compute_jacobian
        )

        # Initial conditions for the states
        self._y0 = self.module.init_state_values()

        # The model parameters
        self._model_params = self.module.init_parameter_values()

        self._ode = ode

    def update_model_parameter(self):
        """
        Update model parameters according
        to parameters in the cell model
        """
        self._model_params = np.array(self._ode.parameter_values(), dtype="float64")

    def solve(self, *args, **kwargs):
        """
        Solver ODE. See docs for the different solvers
        for solver specific arguments.
        """

        self.update_model_parameter()

        # def handler(signum, frame):
        #     raise TimeoutError(('Time to solve ODE has exceeded '
        #                         'the max solving time'))
        # old = signal.signal(signal.SIGALRM, handler)
        # signal.alarm(self.max_solve_time)
        # try:
        #     print('Solving with max solve time = {}'.format(self.max_solve_time))
        ret = self._solve(*args, **kwargs)
        # finally:
        #     signal.signal(signal.SIGALRM, old)
        # signal.alarm(0)
        return ret

    @property
    def module(self):
        return self._module

    def monitor_indices(self, *monitored):
        return self.module.monitor_indices(*monitored)

    @property
    def monitor_names(self):
        return self._monitored

    def state_indices(self, *states):
        return self.module.state_indices(*states)

    @property
    def state_names(self):
        return self._state_names

    def monitor(self, tsteps, results):
        """
        Get monitored values
        """

        monitored_results = np.zeros(
            (len(tsteps), len(self.monitor_names)), dtype=np.float_
        )
        monitored_get_values = np.zeros(len(self.monitor_names), dtype=np.float_)

        for ind, (time, res) in enumerate(zip(tsteps, results)):

            self._eval_monitored(time, res, self._model_params, monitored_get_values)
            monitored_results[ind, :] = monitored_get_values

        return monitored_results

    def _eval_monitored(self, time, res, params, values):
        self.module.monitor(res, time, params, values)


def check_method(method):
    msg = "Unknown method {1}, possible method are {0}".format(method, methods)
    assert method in methods, msg


def generate_module(ode, monitored, **options):
    """
    Generate a module to
    """

    from gotran.codegeneration.compilemodule import compile_module
    from gotran.common.options import parameters

    # Copy of default parameters
    generation = parameters.generation.copy()

    generation.functions.monitored.generate = options.pop("generate_monitored", True)

    generation.code.default_arguments = options.pop("arguments", "tsp")

    generation.functions.rhs.generate = True
    generation.functions.jacobian.generate = options.pop("generate_jacobian", False)

    # Language for the module ("C" or "Python")
    language = options.pop("language", "C")

    additional_declarations = options.pop("additional_declarations", None)
    jacobian_declaration_template = options.pop("jacobian_declaration_template", None)

    module = compile_module(
        ode,
        language,
        monitored,
        generation,
        additional_declarations,
        jacobian_declaration_template,
    )

    return module


# Local imports
from .sundialssolver import SundialsSolver, SundialsNotInstalled
from .scipysolver import ScipySolver


def ODESolver(ode, method="scipy", **options):
    """
    A generic ODE solver for solving problem of the types on the form,

    .. math::

        \dot{y} = f(t,y), \quad y(t_0) = y_0.

    Here one need to specific the backend which is either Scipy or Assimulo.

    *Arguments*

    ode : gotran.ODE or gotran.CellModel
        The ode you want to solve in a gotran object
    method : str
       Solver method. Possible inputs are  or 'scipy' (Default:'sundials')

    options : dict:
       Options for the solver, see `list_solver_options`
    """

    check_method(method)

    if method == "scipy":
        return ScipySolver(ode, **options)
    elif method in sundials_methods:
        try:
            return SundialsSolver(ode, method, **options)
        except:
            print("Could not import Sundials solvers. Use Scipy ODE solver instead")
            return ScipySolver(ode)

    elif method in gotran_methods:
        raise NotImplementedError
    elif method in goss_methods:
        raise NotImplementedError
    else:
        raise NotImplementedError



# Copyright (C) 2013 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    "JacobianComponent",
    "JacobianActionComponent",
    "FactorizedJacobianComponent",
    "ForwardBackwardSubstitutionComponent",
    "LinearizedDerivativeComponent",
    "CommonSubExpressionODE",
    "componentwise_derivative",
    "linearized_derivatives",
    "jacobian_expressions",
    "jacobian_action_expressions",
    "factorized_jacobian_expressions",
    "forward_backward_subst_expressions",
    "diagonal_jacobian_expressions",
    "rhs_expressions",
    "diagonal_jacobian_action_expressions",
    "monitored_expressions",
]

# System imports
import sys

# ModelParameters imports
from modelparameters.sympytools import sp
from modelparameters.codegeneration import sympycode

# Local imports
from gotran.common import (
    error,
    info,
    debug,
    check_arg,
    check_kwarg,
    scalars,
    Timer,
    warning,
    tuplewrap,
    parameters,
    warning,
)
from gotran.common import listwrap
from gotran.model.utils import ode_primitives
from gotran.model.odeobjects import State, Parameter, IndexedObject, Comment
from gotran.model.expressions import *
from gotran.model.ode import ODE
from gotran.codegeneration.codecomponent import CodeComponent

# FIXME: Remove our own cse, or move to this module?
from gotran.codegeneration.sympy_cse import cse


def rhs_expressions(ode, function_name="rhs", result_name="dy", params=None):
    """
    Return a code component with body expressions for the right hand side

    Arguments
    ---------
    ode : gotran.ODE
        The finalized ODE
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the rhs result
    params : dict
        Parameters determining how the code should be generated
    """

    check_arg(ode, ODE)
    if not ode.is_finalized:
        error(
            "Cannot compute right hand side expressions if the ODE is " "not finalized"
        )

    descr = "Compute the right hand side of the {0} ODE".format(ode)

    return CodeComponent(
        "RHSComponent",
        ode,
        function_name,
        descr,
        params=params,
        **{result_name: ode.state_expressions}
    )


def monitored_expressions(
    ode,
    monitored,
    function_name="monitored_expressions",
    result_name="monitored",
    params=None,
):
    """
    Return a code component with body expressions to calculate monitored expressions

    Arguments
    ---------
    ode : gotran.ODE
        The finalized ODE for which the monitored expression should be computed
    monitored : tuple, list
        A tuple/list of strings containing the name of the monitored expressions
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the rhs result
    params : dict
        Parameters determining how the code should be generated
    """

    check_arg(ode, ODE)
    if not ode.is_finalized:
        error(
            "Cannot compute right hand side expressions if the ODE is " "not finalized"
        )

    check_arg(monitored, (tuple, list), itemtypes=str)
    monitored_exprs = []
    for expr_str in monitored:
        obj = ode.present_ode_objects.get(expr_str)
        if not isinstance(obj, Expression):
            error("{0} is not an expression in the {1} ODE".format(expr_str, ode))

        monitored_exprs.append(obj)

    descr = "Computes monitored expressions of the {0} ODE".format(ode)
    return CodeComponent(
        "MonitoredExpressions",
        ode,
        function_name,
        descr,
        params=params,
        **{result_name: monitored_exprs}
    )


def componentwise_derivative(ode, indices, params=None, result_name="dy"):
    """
    Return an ODEComponent holding the expressions for the ith
    state derivative

    Arguments
    ---------
    ode : gotran.ODE
        The finalized ODE for which the ith derivative should be computed
    indices : int, list of ints
        The index
    params : dict
        Parameters determining how the code should be generated
    result_name : str
        The name of the result variable
    """
    check_arg(ode, ODE)
    if not ode.is_finalized:
        error("Cannot compute component wise derivatives if ODE is " "not finalized")
    indices = listwrap(indices)

    check_arg(indices, list, itemtypes=int)
    check_arg(result_name, str)

    if len(indices) == 0:
        error("expected at least on index")

    registered = []
    for index in indices:
        if index < 0 or index >= ode.num_full_states:
            error(
                "Expected the passed indices to be between 0 and the "
                "number of states in the ode, got {0}.".format(index)
            )
        if index in registered:
            error("Index {0} appeared twice.".format(index))

        registered.append(index)

    # Get state expression
    exprs = [ode.state_expressions[index] for index in indices]

    results = {result_name: exprs}

    return CodeComponent(
        "componentwise_derivatives_{0}".format(
            "_".join(expr.state.name for expr in exprs)
        ),
        ode,
        "",
        "",
        params=params,
        **results
    )


def linearized_derivatives(
    ode,
    function_name="linear_derivatives",
    result_names=["linearized", "dy"],
    only_linear=True,
    include_rhs=False,
    nonlinear_last=False,
    params=None,
):
    """
    Return an ODEComponent holding the linearized derivative expressions

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which derivatives should be linearized
    function_name : str
        The name of the function which should be generated
    result_names : str
        The name of the variable storing the linearized derivatives and the
        rhs evaluation if that is included.
    only_linear : bool
        If True, only linear terms will be linearized
    include_rhs : bool
        If True, rhs evaluation will be included in the generated code.
    nonlinear_last : bool
        If True the nonlinear expressions are added last after a comment
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return LinearizedDerivativeComponent(
        ode,
        function_name,
        result_names,
        only_linear,
        include_rhs,
        nonlinear_last,
        params,
    )


def jacobian_expressions(
    ode, function_name="compute_jacobian", result_name="jac", params=None
):
    """
    Return an ODEComponent holding expressions for the jacobian

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the jacobian result
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return JacobianComponent(
        ode, function_name=function_name, result_name=result_name, params=params
    )


def jacobian_action_expressions(
    jacobian,
    with_body=True,
    function_name="compute_jacobian_action",
    result_name="jac_action",
    params=None,
):
    """
    Return an ODEComponent holding expressions for the jacobian action

    Arguments
    ---------
    jacobian : gotran.JacobianComponent
        The ODEComponent holding expressions for the jacobian
    with_body : bool
        If true, the body for computing the jacobian will be included
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the jacobian diagonal result
    params : dict
        Parameters determining how the code should be generated
    """

    check_arg(jacobian, JacobianComponent)
    return JacobianActionComponent(
        jacobian, with_body, function_name, result_name, params=params
    )


def diagonal_jacobian_expressions(
    jacobian,
    function_name="compute_diagonal_jacobian",
    result_name="diag_jac",
    params=None,
):
    """
    Return an ODEComponent holding expressions for the diagonal jacobian

    Arguments
    ---------
    jacobian : gotran.JacobianComponent
        The Jacobian of the ODE
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the jacobian diagonal result
    params : dict
        Parameters determining how the code should be generated
    """
    return DiagonalJacobianComponent(
        jacobian, function_name, result_name, params=params
    )


def diagonal_jacobian_action_expressions(
    diagonal_jacobian,
    with_body=True,
    function_name="compute_diagonal_jacobian_action",
    result_name="diag_jac_action",
    params=None,
):
    """
    Return an ODEComponent holding expressions for the diagonal jacobian action

    Arguments
    ---------
    diagonal_jacobian : gotran.DiagonalJacobianComponent
        The ODEComponent holding expressions for the diagonal jacobian
    with_body : bool
        If true, the body for computing the jacobian will be included
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the variable storing the jacobian diagonal result
    params : dict
        Parameters determining how the code should be generated
    """

    check_arg(diagonal_jacobian, DiagonalJacobianComponent)
    return DiagonalJacobianActionComponent(
        diagonal_jacobian, with_body, function_name, result_name, params=params
    )


def factorized_jacobian_expressions(
    jacobian, function_name="lu_factorize", params=None
):
    """
    Return an ODEComponent holding expressions for the factorized jacobian

    Arguments
    ---------
    jacobian : gotran.JacobianComponent
        The ODEComponent holding expressions for the jacobian
    params : dict
        Parameters determining how the code should be generated
    """
    check_arg(jacobian, JacobianComponent)
    return FactorizedJacobianComponent(jacobian, function_name, params=params)


def forward_backward_subst_expressions(
    factorized,
    function_name="forward_backward_subst",
    result_name="dx",
    residual_name="F",
    params=None,
):
    """
    Return an ODEComponent holding expressions for the forward backward
    substitions for a factorized jacobian

    Arguments
    ---------
    factorized : gotran.FactorizedJacobianComponent
        The ODEComponent holding expressions for the factorized jacobian
    function_name : str
        The name of the function which should be generated
    result_name : str
        The name of the result (increment)
    residual_name : str
        The name of the residual
    params : dict
        Parameters determining how the code should be generated
    """
    check_arg(factorized, FactorizedJacobianComponent)
    return ForwardBackwardSubstitutionComponent(
        factorized,
        function_name=function_name,
        result_name=result_name,
        residual_name=residual_name,
        params=params,
    )


class JacobianComponent(CodeComponent):
    """
    An ODEComponent which keeps all expressions for the Jacobian of the rhs
    """

    def __init__(
        self, ode, function_name="compute_jacobian", result_name="jac", params=None
    ):
        """
        Create a JacobianComponent

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        result_name : str
            The name of the variable storing the jacobian result
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(ode, ODE)

        # Call base class using empty result_expressions
        descr = "Compute the jacobian of the right hand side of the " "{0} ODE".format(
            ode
        )
        super(JacobianComponent, self).__init__(
            "Jacobian", ode, function_name, descr, params=params
        )
        check_arg(result_name, str)

        timer = Timer("Computing jacobian")

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states

        # Create Jacobian matrix
        N = len(states)
        self.jacobian = sp.Matrix(N, N, lambda i, j: 0.0)

        self.num_nonzero = 0

        self.shapes[result_name] = (N, N)

        state_dict = dict((state.sym, ind) for ind, state in enumerate(states))
        time_sym = states[0].time.sym

        might_take_time = N >= 10

        if might_take_time:
            info(
                "Calculating Jacobian of {0}. Might take some time...".format(ode.name)
            )
            sys.stdout.flush()

        for i, expr in enumerate(state_exprs):

            states_syms = sorted(
                (state_dict[sym], sym)
                for sym in ode_primitives(expr.expr, time_sym)
                if sym in state_dict
            )

            self.add_comment(
                "Expressions for the sparse jacobian of "
                "state {0}".format(expr.state.name),
                dependent=expr,
            )

            for j, sym in states_syms:
                time_diff = Timer("Differentiate state_expressions")
                jac_ij = expr.expr.diff(sym)
                del time_diff
                self.num_nonzero += 1
                jac_ij = self.add_indexed_expression(
                    result_name, (i, j), jac_ij, dependent=expr
                )

                self.jacobian[i, j] = jac_ij

        if might_take_time:
            info(" done")

        # Call recreate body with the jacobian expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class DiagonalJacobianComponent(CodeComponent):
    """
    An ODEComponent which keeps all expressions for the Jacobian of the rhs
    """

    def __init__(
        self,
        jacobian,
        function_name="compute_diagonal_jacobian",
        result_name="diag_jac",
        params=None,
    ):
        """
        Create a DiagonalJacobianComponent

        Arguments
        ---------
        jacobian : gotran.JacobianComponent
            The Jacobian of the ODE
        function_name : str
            The name of the function which should be generated
        result_name : str (optional)
            The basename of the indexed result expression
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(jacobian, JacobianComponent)

        descr = (
            "Compute the diagonal jacobian of the right hand side of the "
            "{0} ODE".format(jacobian.root)
        )
        super(DiagonalJacobianComponent, self).__init__(
            "DiagonalJacobian", jacobian.root, function_name, descr, params=params
        )

        what = "Computing diagonal jacobian"
        timer = Timer(what)

        self.add_comment(what)

        N = jacobian.jacobian.shape[0]
        self.shapes[result_name] = (N,)
        jacobian_name = jacobian.results[0]

        # Create IndexExpressions of the diagonal Jacobian
        for expr in jacobian.indexed_objects(jacobian_name):
            if expr.indices[0] == expr.indices[1]:
                self.add_indexed_expression(result_name, expr.indices[0], expr.expr)

        self.diagonal_jacobian = sp.Matrix(N, N, lambda i, j: 0.0)

        for i in range(N):
            self.diagonal_jacobian[i, i] = jacobian.jacobian[i, i]

        # Call recreate body with the jacobian diagonal expressions as the
        # result expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class JacobianActionComponent(CodeComponent):
    """
    Jacobian action component which returns the expressions for Jac*x
    """

    def __init__(
        self,
        jacobian,
        with_body=True,
        function_name="compute_jacobian_action",
        result_name="jac_action",
        params=None,
    ):
        """
        Create a JacobianActionComponent

        Arguments
        ---------
        jacobian : gotran.JacobianComponent
            The Jacobian of the ODE
        with_body : bool
            If true, the body for computing the jacobian will be included
        function_name : str
            The name of the function which should be generated
        result_name : str
            The basename of the indexed result expression
        params : dict
            Parameters determining how the code should be generated
        """
        timer = Timer("Computing jacobian action component")
        check_arg(jacobian, JacobianComponent)
        descr = (
            "Compute the jacobian action of the right hand side of the "
            "{0} ODE".format(jacobian.root)
        )
        super(JacobianActionComponent, self).__init__(
            "JacobianAction", jacobian.root, function_name, descr, params=params
        )

        x = self.root.full_state_vector
        jac = jacobian.jacobian
        jacobian_name = jacobian.results[0]

        # Create Jacobian action vector
        self.action_vector = sp.Matrix(len(x), 1, lambda i, j: 0)

        self.add_comment("Computing the action of the jacobian")

        self.shapes[result_name] = (len(x),)
        self.shapes[jacobian_name] = jacobian.shapes[jacobian_name]
        for i, expr in enumerate(jac * x):
            self.action_vector[i] = self.add_indexed_expression(result_name, i, expr)

        # Call recreate body with the jacobian action expressions as the
        # result expressions
        results = {result_name: self.indexed_objects(result_name)}
        if with_body:
            results, body_expressions = self._body_from_results(**results)
        else:
            body_expressions = results[result_name]

        self.body_expressions = self._recreate_body(body_expressions, **results)


class DiagonalJacobianActionComponent(CodeComponent):
    """
    Jacobian action component which returns the expressions for Jac*x
    """

    def __init__(
        self,
        diagonal_jacobian,
        with_body=True,
        function_name="compute_diagonal_jacobian_action",
        result_name="diag_jac_action",
        params=None,
    ):
        """
        Create a DiagonalJacobianActionComponent

        Arguments
        ---------
        jacobian : gotran.JacobianComponent
            The Jacobian of the ODE
        with_body : bool
            If true, the body for computing the jacobian will be included
        function_name : str
            The name of the function which should be generated
        result_name : str
            The basename of the indexed result expression
        params : dict
            Parameters determining how the code should be generated
        """
        timer = Timer("Computing jacobian action component")
        check_arg(diagonal_jacobian, DiagonalJacobianComponent)
        descr = (
            "Compute the diagonal jacobian action of the right hand side "
            "of the {0} ODE".format(diagonal_jacobian.root)
        )
        super(DiagonalJacobianActionComponent, self).__init__(
            "DiagonalJacobianAction",
            diagonal_jacobian.root,
            function_name,
            descr,
            params=params,
        )

        x = self.root.full_state_vector
        jac = diagonal_jacobian.diagonal_jacobian

        self._action_vector = sp.Matrix(len(x), 1, lambda i, j: 0)

        self.add_comment("Computing the action of the jacobian")

        # Create Jacobian matrix
        self.shapes[result_name] = (len(x),)
        for i, expr in enumerate(jac * x):
            self._action_vector[i] = self.add_indexed_expression(result_name, i, expr)

        # Call recreate body with the jacobian action expressions as the
        # result expressions
        results = {result_name: self.indexed_objects(result_name)}
        if with_body:
            results, body_expressions = self._body_from_results(**results)
        else:
            body_expressions = results[result_name]

        self.body_expressions = self._recreate_body(body_expressions, **results)


class FactorizedJacobianComponent(CodeComponent):
    """
    Class to generate expressions for symbolicaly factorizing a jacobian
    """

    def __init__(self, jacobian, function_name="lu_factorize", params=None):
        """
        Create a FactorizedJacobianComponent

        Arguments
        ---------
        jacobian : gotran.JacobianComponent
            The Jacobian of the ODE
        function_name : str
            The name of the function which should be generated
        params : dict
            Parameters determining how the code should be generated
        """

        timer = Timer("Computing factorization of jacobian")
        check_arg(jacobian, JacobianComponent)
        descr = "Symbolically factorize the jacobian of the {0} ODE".format(
            jacobian.root
        )
        super(FactorizedJacobianComponent, self).__init__(
            "FactorizedJacobian",
            jacobian.root,
            function_name,
            descr,
            params=params,
            use_default_arguments=False,
            additional_arguments=jacobian.results,
        )

        self.add_comment("Factorizing jacobian of {0}".format(self.root.name))

        jacobian_name = jacobian.results[0]

        # Recreate jacobian using only sympy Symbols
        jac_orig = jacobian.jacobian

        # Size of system
        n = jac_orig.rows
        jac = sp.Matrix(n, n, lambda i, j: sp.S.Zero)

        for i in range(n):
            for j in range(n):
                # print jac_orig[i,j]
                if not jac_orig[i, j].is_zero:
                    name = sympycode(jac_orig[i, j])
                    jac[i, j] = sp.Symbol(
                        name,
                        real=True,
                        imaginary=False,
                        commutative=True,
                        hermitian=True,
                        complex=True,
                    )
                    print(jac[i, j])
        p = []

        self.shapes[jacobian_name] = (n, n)

        def add_intermediate_if_changed(jac, jac_ij, i, j):
            # If item has changed
            if jac_ij != jac[i, j]:
                print("jac", i, j, jac_ij)
                jac[i, j] = self.add_indexed_expression(jacobian_name, (i, j), jac_ij)

        # Do the factorization
        for j in range(n):

            for i in range(j):

                # Get sympy expr of A_ij
                jac_ij = jac[i, j]

                # Build sympy expression
                for k in range(i):
                    jac_ij -= jac[i, k] * jac[k, j]

                add_intermediate_if_changed(jac, jac_ij, i, j)

            pivot = -1

            for i in range(j, n):

                # Get sympy expr of A_ij
                jac_ij = jac[i, j]

                # Build sympy expression
                for k in range(j):
                    jac_ij -= jac[i, k] * jac[k, j]

                add_intermediate_if_changed(jac, jac_ij, i, j)

                # find the first non-zero pivot, includes any expression
                if pivot == -1 and jac[i, j]:
                    pivot = i

            if pivot < 0:
                # this result is based on iszerofunc's analysis of the
                # possible pivots, so even though the element may not be
                # strictly zero, the supplied iszerofunc's evaluation gave
                # True
                error("No nonzero pivot found; symbolic inversion failed.")

            if pivot != j:  # row must be swapped
                jac.row_swap(pivot, j)
                p.append([pivot, j])
                print("Pivoting!!")

            # Scale with diagonal
            if not jac[j, j]:
                error("Diagonal element of the jacobian is zero. " "Inversion failed")

            scale = 1 / jac[j, j]
            for i in range(j + 1, n):

                # Get sympy expr of A_ij
                jac_ij = jac[i, j]
                jac_ij *= scale
                add_intermediate_if_changed(jac, jac_ij, i, j)

        # Store factorized jacobian
        self.factorized_jacobian = jac
        self.num_nonzero = sum(
            not jac[i, j].is_zero for i in range(n) for j in range(n)
        )

        # No need to call recreate body expressions
        self.body_expressions = self.ode_objects

        self.used_states = set()
        self.used_parameters = set()


class ForwardBackwardSubstitutionComponent(CodeComponent):
    """
    Class to generate a forward backward substiution algorithm for
    symbolically factorized jacobian
    """

    def __init__(
        self,
        factorized,
        function_name="forward_backward_subst",
        result_name="dx",
        residual_name="F",
        params=None,
    ):
        """
        Create a JacobianForwardBackwardSubstComponent

        Arguments
        ---------
        factorized : gotran.FactorizedJacobianComponent
            The factorized jacobian of the ODE
        function_name : str
            The name of the function which should be generated
        result_name : str
            The name of the result (increment)
        residual_name : str
            The name of the residual
        params : dict
            Parameters determining how the code should be generated
        """
        timer = Timer("Computing forward backward substituion component")
        check_arg(factorized, FactorizedJacobianComponent)
        jacobian_name = list(factorized.shapes.keys())[0]
        additional_indexed_arguments = factorized.additional_arguments + [residual_name]
        descr = (
            "Symbolically forward backward substitute linear system "
            "of {0} ODE".format(factorized.root)
        )
        super(ForwardBackwardSubstitutionComponent, self).__init__(
            "ForwardBackwardSubst",
            factorized.root,
            function_name,
            descr,
            params=params,
            use_default_arguments=False,
            additional_arguments=[residual_name],
        )

        self.add_comment(
            "Forward backward substituting factorized "
            "linear system {0}".format(self.root.name)
        )

        # Recreate jacobian using only sympy Symbols
        jac_orig = factorized.factorized_jacobian

        # Size of system
        n = jac_orig.rows
        jac = sp.Matrix(n, n, lambda i, j: sp.S.Zero)

        for i in range(n):
            for j in range(n):
                # print jac_orig[i,j]
                if not jac_orig[i, j].is_zero:
                    name = sympycode(jac_orig[i, j])
                    jac[i, j] = sp.Symbol(
                        name,
                        real=True,
                        imaginary=False,
                        commutative=True,
                        hermitian=True,
                        complex=True,
                    )
                    print(jac[i, j])

        self.shapes[jacobian_name] = (n, n)
        self.shapes[residual_name] = (n,)
        self.shapes[result_name] = (n,)

        F = []
        dx = []
        # forward substitution, all diag entries are scaled to 1
        for i in range(n):

            F.append(self.add_indexed_object(residual_name, i))
            dx.append(self.add_indexed_expression(result_name, i, F[i]))

            for j in range(i):
                if jac[i, j].is_zero:
                    continue
                dx[i] = self.add_indexed_expression(
                    result_name, i, dx[i] - dx[j] * jac[i, j]
                )

        # backward substitution
        for i in range(n - 1, -1, -1):
            for j in range(i + 1, n):
                if jac[i, j].is_zero:
                    continue
                dx[i] = self.add_indexed_expression(
                    result_name, i, dx[i] - dx[j] * jac[i, j]
                )

            dx[i] = self.add_indexed_expression(result_name, i, dx[i] / jac[i, i])

        # No need to call recreate body expressions
        self.body_expressions = [
            obj
            for obj in self.ode_objects
            if isinstance(obj, (IndexedExpression, Comment))
        ]

        self.results = [result_name]
        self.used_states = set()
        self.used_parameters = set()


class LinearizedDerivativeComponent(CodeComponent):
    """
    A component for all linear and linearized derivatives
    """

    def __init__(
        self,
        ode,
        function_name="linear_derivatives",
        result_names=["linearized", "rhs"],
        only_linear=True,
        include_rhs=False,
        nonlinear_last=False,
        params=None,
    ):
        """
        Return an ODEComponent holding the linearized derivative expressions

        Arguments
        ---------
        ode : gotran.ODE
            The ODE for which derivatives should be linearized
        function_name : str
            The name of the function which should be generated
        result_names : str
            The name of the variable storing the linearized derivatives and the
            rhs evaluation if that is included.
        only_linear : bool
            If True, only linear terms will be linearized
        include_rhs : bool
            If True, rhs evaluation will be included in the generated code.
        params : dict
            Parameters determining how the code should be generated
        """

        check_kwarg(result_names, "result_names", list, itemtypes=str)
        if len(result_names) != 2:
            error("expected the length of 'result_names' to be 2")

        descr = "Computes the linearized derivatives for all linear derivatives"
        super(LinearizedDerivativeComponent, self).__init__(
            "LinearizedDerivatives", ode, function_name, descr, params=params
        )

        check_arg(ode, ODE)
        assert ode.is_finalized

        self.linear_derivative_indices = [0] * self.root.num_full_states
        linearized_name, rhs_name = result_names

        state_exprs = self.root.state_expressions
        self.shapes[linearized_name] = (self.root.num_full_states,)
        if include_rhs:
            self.shapes[rhs_name] = (self.root.num_full_states,)

        nonlinear_exprs = []
        for ind, expr in enumerate(state_exprs):

            expr_diff = expr.expr.diff(expr.state.sym)

            if expr_diff and expr.state.sym not in expr_diff.args:
                self.linear_derivative_indices[ind] = 1
            elif only_linear:
                continue

            # Append nonlinear expressions for later addition
            if nonlinear_last and self.linear_derivative_indices[ind] == 0:
                nonlinear_exprs.append((linearized_name, ind, expr_diff))
            else:
                self.add_indexed_expression(
                    linearized_name, ind, expr_diff, dependent=expr
                )

        if nonlinear_exprs:
            self.add_comment("Nonlinear linearized expressions", dependent=expr)
            for linearized_name, ind, expr_diff in nonlinear_exprs:
                self.add_indexed_expression(
                    linearized_name, ind, expr_diff, dependent=expr
                )

        # Call recreate body with the jacobian action expressions as the
        # result expressions
        results = {linearized_name: self.indexed_objects(linearized_name)}
        if include_rhs:
            results[rhs_name] = state_exprs
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class CommonSubExpressionODE(ODE):
    """
    Class which flattens the component structue of an ODE to just one.
    It uses common sub expressions as intermediates to reduce complexity
    of the derivative expressions.
    """

    def __init__(self, ode):
        check_arg(ode, ODE)
        assert ode.is_finalized

        timer = Timer("Extract common sub expressions")

        newname = ode.name + "_CSE"

        # Call super class
        super(CommonSubExpressionODE, self).__init__(newname, ode.ns)

        # Add states and parameters
        atoms = []
        for state in ode.full_states:
            atoms.append(self.add_state(state.name, state.param))

        for param in ode.parameters:
            atoms.append(self.add_parameter(param.name, param.param))

        # Collect all expanded state expressions
        org_state_expressions = ode.state_expressions
        expanded_state_exprs = [
            ode.expanded_expressions[obj.name] for obj in org_state_expressions
        ]

        # Call sympy common sub expression reduction
        cse_exprs, cse_state_exprs = cse(
            expanded_state_exprs, symbols=sp.numbered_symbols("cse_"), optimizations=[]
        )
        cse_cnt = 0
        cse_subs = {}

        # Register the common sub expressions as Intermediates
        for sub, expr in cse_exprs:

            # If the expression is just one of the atoms of the ODE we skip
            # the cse expressions but add a subs for the atom
            if expr in atoms:
                cse_subs[sub] = expr
            else:
                cse_subs[sub] = self.add_intermediate(
                    "cse_{0}".format(cse_cnt), expr.xreplace(cse_subs)
                )
                cse_cnt += 1

        # Register the state expressions
        for org_state_expr, state_expr in zip(org_state_expressions, cse_state_exprs):

            exp_expr = state_expr.xreplace(cse_subs)
            state = self.get_object(org_state_expr.state.name)[1]

            # If state derivative
            if isinstance(org_state_expr, StateDerivative):
                self.add_derivative(state, state.time.sym, exp_expr)

            # If algebraic
            elif isinstance(org_state_expr, AlgebraicExpression):
                self.add_algebraic(state, exp_expr)

            else:
                error("Should not come here!")

        self.finalize()



# Copyright (C) 2013-2014 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    "ode_primitives",
    "INTERMEDIATE",
    "ALGEBRAIC_EXPRESSION",
    "DERIVATIVE_EXPRESSION",
    "STATE_SOLUTION_EXPRESSION",
    "special_expression",
    "iter_objects",
    "ode_objects",
    "ode_components",
    "ODEObjectList",
    "RateDict",
]

# System imports
from collections import OrderedDict
import weakref
import re

from sympy.core.function import AppliedUndef
from sympy import preorder_traversal, Symbol

# ModelParameters imports
from modelparameters.sympytools import sp
from modelparameters.utils import tuplewrap

# Local imports
from gotran.common import error, debug, check_arg, check_kwarg, scalars
from gotran.model.odeobjects import *
from gotran.model.expressions import *


def ode_primitives(expr, time):
    """
    Return all ODE primitives

    Arguments
    ---------
    expr : sympy.expression
        A sympy expression of ode symbols
    time : sympy.Symbol
        A Symbol representing time in the ODE
    """
    symbols = set()
    pt = preorder_traversal(expr)

    for node in pt:

        # Collect AppliedUndefs which are functions of time
        if (
            isinstance(node, AppliedUndef)
            and len(node.args) == 1
            and node.args[0] == time
        ):
            pt.skip()
            symbols.add(node)
        elif isinstance(node, Symbol):
            symbols.add(node)

    return symbols


_derivative_name_template = re.compile("\Ad([a-zA-Z]\w*)_d([a-zA-Z]\w*)\Z")
_algebraic_name_template = re.compile("\Aalg_([a-zA-Z]\w*)_0\Z")

# Flags for determine special expressions
INTERMEDIATE = 0
ALGEBRAIC_EXPRESSION = 1
DERIVATIVE_EXPRESSION = 2
STATE_SOLUTION_EXPRESSION = 3


def special_expression(name, root):
    """
    Check if an expression name corresponds to a special expression
    """

    alg_expr = re.search(_algebraic_name_template, name)
    if alg_expr:
        return alg_expr, ALGEBRAIC_EXPRESSION

    der_expr = re.search(_derivative_name_template, name)
    if der_expr:
        return der_expr, DERIVATIVE_EXPRESSION

    state = root.present_ode_objects.get(name)
    if state and isinstance(state, State):
        return state, STATE_SOLUTION_EXPRESSION

    return None, INTERMEDIATE


class iter_objects(object):
    """
    A recursive iterator over all objects of a component including its
    childrens

    Arguments
    ---------
    comp : gotran.ODEComponent
        The root ODEComponent of the iteration
    reverse : bool
        If True the iteration is done from the last component added
    types : gotran.ODEObject types (optional)
        Only iterate over particular types

    Yields
    ------
    ode_object : gotran.ODEObject
        All ODEObjects of a component
    """

    def __init__(
        self, comp, return_comp=True, only_return_comp=False, reverse=False, *types
    ):
        from .odecomponent import ODEComponent

        assert isinstance(comp, ODEComponent)
        self._types = tuplewrap(types) or (ODEObject,)
        self._return_comp = return_comp
        self._only_return_comp = only_return_comp
        if reverse:
            self._object_iterator = self._reverse_iter_objects(comp)
        else:
            self._object_iterator = self._iter_objects(comp)

        assert all(issubclass(T, ODEObject) for T in self._types)

    def _reverse_iter_objects(self, comp):

        # First all children components in reversed order
        for sub_comp in reversed(list(comp.children.values())):
            for sub_tree in self._reverse_iter_objects(sub_comp):
                yield sub_tree

        # Secondly return component
        if self._return_comp or self._only_return_comp:
            yield comp

        if self._only_return_comp:
            return

        # Last all objects
        for obj in reversed(comp.ode_objects):
            if isinstance(obj, self._types):
                yield obj

    def _iter_objects(self, comp):

        # First return component
        if self._return_comp:
            yield comp

        # Secondly all objects
        if not self._only_return_comp:

            for obj in comp.ode_objects:
                if isinstance(obj, self._types):
                    yield obj

        # Thrirdly all children components
        for sub_comp in list(comp.children.values()):
            for sub_tree in self._iter_objects(sub_comp):
                yield sub_tree

    def __next__(self):
        return next(self._object_iterator)

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()


def ode_objects(comp, *types):
    """
    Return a list of ode objects

    Arguments
    ---------
    comp : gotran.ODEComponent
        The root ODEComponent of the list
    types : gotran.ODEObject types (optional)
        Only include objects of type given in types
    """
    return [obj for obj in iter_objects(comp, False, False, False, *types)]


def ode_components(comp, include_self=True):
    """
    Return a list of ode components

    Arguments
    ---------
    comp : gotran.ODEComponent
        The root ODEComponent of the list
    return_self : bool (optional)
        The list will include the passed component if True
    """
    comps = [obj for obj in iter_objects(comp, True, True)]

    if not include_self:
        comps.remove(comp)

    return comps


class ODEObjectList(list):
    """
    Specialized container for ODEObjects. It is a list but adds dict
    access through the name attribute of an ODEObjects
    """

    def __init__(self):
        """
        Initialize ODEObjectList. Only empty such.
        """
        super(ODEObjectList, self).__init__()
        self._objects = {}

    def keys(self):
        return list(self._objects.keys())

    def append(self, item):
        check_arg(item, ODEObject, 0, ODEObjectList.append)
        super(ODEObjectList, self).append(item)
        self._objects[item.name] = item

    def insert(self, index, item):
        check_arg(item, ODEObject, 1, ODEObjectList.insert)
        super(ODEObjectList, self).insert(index, item)
        self._objects[item.name] = item

    def extend(self, iterable):
        check_arg(iterable, list, 0, ODEObjectList.extend, ODEObject)
        super(ODEObjectList, self).extend(iterable)
        for item in iterable:
            self._objects[item.name] = item

    def get(self, name):
        if isinstance(name, str):
            return self._objects.get(name)
        elif isinstance(name, sp.Symbol):
            return self._objects.get(name.name)
        return None

    def __contains__(self, item):
        if isinstance(item, str):
            return any(item == obj.name for obj in self)
        elif isinstance(item, sp.Symbol):
            return any(item.name == obj.name for obj in self)
        elif (item, ODEObject):
            return super(ODEObjectList, self).__contains__(item)
        return False

    def count(self, item):
        if isinstance(item, str):
            return sum(item == obj.name for obj in self)
        elif isinstance(item, sp.Symbol):
            return sum(item.name == obj.name for obj in self)
        elif (item, ODEObject):
            return super(ODEObjectList, self).count(item)
        return 0

    def index(self, item):
        if isinstance(item, str):
            for ind, obj in enumerate(self):
                if item == obj.name:
                    return ind
        elif isinstance(item, sp.Symbol):
            for ind, obj in enumerate(self):
                if item.name == obj.name:
                    return ind
        elif (item, ODEObject):
            for ind, obj in enumerate(self):
                if item == obj:
                    return ind
        raise ValueError("Item '{0}' not part of this ODEObjectList.".format(str(item)))

    def sort(self):
        error("Cannot sort ODEObjectList.")

    def pop(self, index):

        check_arg(index, int)
        if index >= len(self):
            raise IndexError("pop index out of range")
        obj = super(ODEObjectList, self).pop(index)
        self._objects.pop(obj.name)

    def remove(self, item):
        try:
            index = self.index(item)
        except ValueError:
            raise ValueError("ODEObjectList.remove(x): x not in list")

        self.pop(index)

    def reverse(self, item):
        error("Cannot alter ODEObjectList, other than adding ODEObjects.")


class RateDict(OrderedDict):
    """
    A storage class for Markov model rates
    """

    def __init__(self, comp):
        from .odecomponent import ODEComponent

        check_arg(comp, ODEComponent)
        self._comp = weakref.ref(comp)
        super(RateDict, self).__init__()

    def __setitem__(self, states, expr):
        """
        Register rate(s) between states

        Arguments
        ---------
        states : tuple of two states, list of States, tuple of two lists of States
            If tuple of two states is given a single rate is expected.
            If one list is passed the rate expression should be a square matrix
            and the states list determines the order of the row and column of
            the matrix. If two lists are passed the first determines the states
            in the row and the second the states in the column of the Matrix
        expr : sympy.Basic, scalar, sympy.MatrixBase
            A sympy.Basic and scalars is expected for a single rate between
            two states.
            A sympy.Matrix is expected if several rate expressions are given,
            see explaination in for states argument for how the columns and
            rows are interpreted.
        """

        if isinstance(expr, sp.Matrix):
            self._comp()._add_rates(states, expr)
        else:
            if not isinstance(states, tuple) or len(states) != 2:
                error(
                    "Expected a tuple of size 2 with states when "
                    "registering a single rate."
                )

            # NOTE: the actuall item is set by the component while calling this
            # function, using _register_single_rate. See below.
            self._comp()._add_single_rate(states[0], states[1], expr)

    def _register_single_rate(self, to_state, from_state, expr_sym):
        """
        Actually setting an item
        """
        OrderedDict.__setitem__(self, (to_state, from_state), expr_sym)



# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

from . import odeobjects
from . import expressions
from . import odecomponent
from . import utils
from . import ode
from . import cellmodel
from . import loadmodel


# gotran imports
# from .odeobjects import *
from .odeobjects import (
    ODEObject,
    State,
    ODEValueObject,
    Parameter,
    Comment,
    IndexedObject,
    Time,
    Dt,
    SingleODEObjects,
    cmp,
    cmp_to_key,
)

from .odecomponent import ODEComponent
from .expressions import (
    Expression,
    DerivativeExpression,
    AlgebraicExpression,
    StateExpression,
    StateSolution,
    RateExpression,
    Intermediate,
    StateDerivative,
    Derivatives,
    IndexedExpression,
    recreate_expression,
)
from .ode import ODE
from .cellmodel import CellModel
from .loadmodel import load_ode, exec_ode, load_cell

# __all__ = [_name for _name in list(globals().keys()) if _name[0] != "_"]
__all__ = [
    "odeobjects",
    "ODEObject",
    "Comment",
    "ODEValueObject",
    "Parameter",
    "State",
    "SingleODEObjects",
    "Time",
    "Dt",
    "IndexedObject",
    "cmp_to_key",
    "cmp",
    "odecomponent",
    "ODEComponent",
    "expressions",
    "Expression",
    "DerivativeExpression",
    "AlgebraicExpression",
    "StateExpression",
    "StateSolution",
    "RateExpression",
    "Intermediate",
    "StateDerivative",
    "Derivatives",
    "IndexedExpression",
    "recreate_expression",
    "ode",
    "ODE",
    "cellmodel",
    "CellModel",
    "loadmodel",
    "load_ode",
    "exec_ode",
    "load_cell",
]
